__all__ = ('MessageSnapshot', )

from scarletio import RichAttributeErrorBaseType, export

from ...bases import id_sort_key
from ...core import CHANNELS
from ...role import create_partial_role_from_id
from ...utils import CHANNEL_MENTION_RP, DATETIME_FORMAT_CODE, DISCORD_EPOCH_START

from ..message import MessageFlag, MessageType
from ..message.constants import MESSAGE_STATE_MASK_CACHE_MENTIONED_CHANNELS

from .fields import (
    parse_attachments, parse_components, parse_content, parse_created_at, parse_edited_at, parse_embeds, parse_flags,
    parse_mentioned_role_ids, parse_mentioned_users, parse_soundboard_sounds, parse_stickers, parse_type,
    put_attachments, put_components, put_content, put_created_at, put_edited_at,
    put_embeds, put_flags, put_mentioned_role_ids, put_mentioned_users, put_soundboard_sounds,
    put_stickers, put_type, validate_attachments, validate_components, validate_content, validate_created_at,
    validate_edited_at, validate_embeds, validate_flags, validate_mentioned_role_ids, validate_mentioned_users,
    validate_soundboard_sounds, validate_stickers, validate_type
)


@export
class MessageSnapshot(RichAttributeErrorBaseType):
    """
    Might be sent with a ``Message``, if it has rich presence-related chat embeds.
    
    Attributes
    ----------
    _cache_mentioned_channels : `None`, `tuple` of ``Channel``
        Mentioned channels by the message. Parsed from ``.content``. Defaults to `None`.
        
        Cache field used by ``.mentioned_channels``.
    
    _state : `int`
        Bitwise mask used to track the message's state.
    
    attachments : `None | tuple<Attachment>`
        The snapshotted message's attachments.
    
    components : `None | tuple<Component>`
        The snapshotted message's components.
    
    content : `None | str`
        The snapshotted message's content.
    
    created_at : `DateTime`
        When the snapshotted message was created.
    
    edited_at : `None | Datetime`
        When the snapshotted message was edited.
    
    embeds : `None | tuple<Embed>`
        The snapshotted message's embeds.
    
    flags : ``MessageFlag``
        The snapshotted message's flags.
    
    mentioned_role_ids : `None | tuple<int>`
        The mentioned roles' identifiers.
    
    mentioned_users : `None | tuple<ClientUserBase>`
        The mentioned users.
    
    soundboard_sounds : `None | tuple<SoundboardSound>`
        The soundboard sounds attached to the message. 
    
    stickers : `None | tuple<Sticker>`
        The snapshotted message's stickers.
    
    type : ``MessageType``
        The snapshotted message's type.
    """
    __slots__ = (
        '_cache_mentioned_channels', '_state', 'attachments', 'components', 'content', 'created_at', 'edited_at',
        'embeds', 'flags', 'mentioned_role_ids', 'mentioned_users', 'soundboard_sounds', 'stickers', 'type'
    )
    
    def __new__(
        cls,
        *,
        attachments = ...,
        components = ...,
        content = ...,
        created_at = ...,
        edited_at = ...,
        embeds = ...,
        flags = ...,
        mentioned_role_ids = ...,
        mentioned_users = ...,
        message_type = ...,
        soundboard_sounds = ...,
        stickers = ...,
    ):
        """
        Creates a new message snapshot from the given parameters.
        
        Parameters
        ----------
        attachments : `None | iterable<Attachment>`, Optional (Keyword only)
            The snapshotted message's attachments.
        
        components : `None | iterable<Component>`, Optional (Keyword only)
            The snapshotted message's components.
        
        content : `None | str`, Optional (Keyword only)
            The snapshotted message's content.
        
        created_at : `DateTime`, Optional (Keyword only)
            When the snapshotted message was created.
        
        edited_at : `None | Datetime`, Optional (Keyword only)
            When the snapshotted message was edited.
        
        embeds : `None | iterable<Embed>`, Optional (Keyword only)
            The snapshotted message's embeds.
        
        flags : `MessageFlag | int | None`, Optional (Keyword only)
            The snapshotted message's flags.
        
        mentioned_role_ids : `None | iterable<int>` | iterable<Role>`, Optional (Keyword only)
            The mentioned roles' identifiers.
        
        mentioned_users : `None | iterable<ClientUserBase>`, Optional (Keyword only)
            The mentioned users.
        
        message_type : `MessageType | int | None`, Optional (Keyword only)
            The snapshotted message's type.
        
        soundboard_sounds : `None | iterable<SoundboardSound>`, Optional (Keyword only)
            Soundboard sounds attached to the message.
        
        stickers : `None | iterable<Sticker>`, Optional (Keyword only)
            The snapshotted message's stickers.
        
        Raises
        ------
        TypeError
            - If a parameter's content is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # attachments
        if attachments is ...:
            attachments = None
        else:
            attachments = validate_attachments(attachments)
        
        # components
        if components is ...:
            components = None
        else:
            components = validate_components(components)
        
        # content
        if content is ...:
            content = None
        else:
            content = validate_content(content)
        
        # created_at
        if created_at is ...:
            created_at = DISCORD_EPOCH_START
        else:
            created_at = validate_created_at(created_at)
        
        # edited_at
        if edited_at is ...:
            edited_at = None
        else:
            edited_at = validate_edited_at(edited_at)
        
        # embeds
        if embeds is ...:
            embeds = None
        else:
            embeds = validate_embeds(embeds)
        
        # flags
        if flags is ...:
            flags = MessageFlag()
        else:
            flags = validate_flags(flags)
        
        # mentioned_role_ids
        if mentioned_role_ids is ...:
            mentioned_role_ids = None
        else:
            mentioned_role_ids = validate_mentioned_role_ids(mentioned_role_ids)
        
        # mentioned_users
        if mentioned_users is ...:
            mentioned_users = None
        else:
            mentioned_users = validate_mentioned_users(mentioned_users)
        
        # soundboard_sounds
        if soundboard_sounds is ...:
            soundboard_sounds = None
        else:
            soundboard_sounds = validate_soundboard_sounds(soundboard_sounds)
        
        # stickers
        if stickers is ...:
            stickers = None
        else:
            stickers = validate_stickers(stickers)
        
        # type
        if message_type is ...:
            message_type = MessageType.default
        else:
            message_type = validate_type(message_type)
        
        # Construct
        self = object.__new__(cls)
        self._cache_mentioned_channels = None
        self._state = 0
        self.attachments = attachments
        self.components = components
        self.content = content
        self.created_at = created_at
        self.edited_at = edited_at
        self.embeds = embeds
        self.flags = flags
        self.mentioned_role_ids = mentioned_role_ids
        self.mentioned_users = mentioned_users
        self.soundboard_sounds = soundboard_sounds
        self.stickers = stickers
        self.type = message_type
        return self
    
    
    @classmethod
    def from_data(cls, data, guild_id = 0):
        """
        Creates a new message snapshot from the given data.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Message snapshot data.
        guild_id : `int` = `0`, Optional
            The respective guild's identifier.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self._cache_mentioned_channels = None
        self._state = 0
        self.attachments = parse_attachments(data)
        self.components = parse_components(data)
        self.content = parse_content(data)
        self.created_at = parse_created_at(data)
        self.edited_at = parse_edited_at(data)
        self.embeds = parse_embeds(data)
        self.flags = parse_flags(data)
        self.mentioned_role_ids = parse_mentioned_role_ids(data)
        self.mentioned_users = parse_mentioned_users(data, guild_id)
        self.soundboard_sounds = parse_soundboard_sounds(data)
        self.stickers = parse_stickers(data)
        self.type = parse_type(data)
        return self
    
    
    def to_data(self, *, defaults = False, guild_id = 0):
        """
        Converts the message snapshot back to json a serializable dictionary.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        
        guild_id : `int` = `0`, Optional (Keyword only)
            The respective guild's identifier.
        
        Returns
        -------
        data : `dict<str, object>`
        """
        data = {}
        put_attachments(self.attachments, data, defaults)
        put_components(self.components, data, defaults)
        put_content(self.content, data, defaults)
        put_created_at(self.created_at, data, defaults)
        put_edited_at(self.edited_at, data, defaults)
        put_embeds(self.embeds, data, defaults)
        put_flags(self.flags, data, defaults)
        put_mentioned_role_ids(self.mentioned_role_ids, data, defaults)
        put_mentioned_users(self.mentioned_users, data, defaults, guild_id = guild_id)
        put_soundboard_sounds(self.soundboard_sounds, data, defaults)
        put_stickers(self.stickers, data, defaults)
        put_type(self.type, data, defaults)
        return data
    
    
    def __eq__(self, other):
        """Returns whether the two message activities are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # attachments
        if self.attachments != other.attachments:
            return False
        
        # components
        if self.components != other.components:
            return False
        
        # content
        if self.content != other.content:
            return False
        
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # edited_at
        if self.edited_at != other.edited_at:
            return False
        
        # embeds
        if self.embeds != other.embeds:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # mentioned_role_ids
        if self.mentioned_role_ids != other.mentioned_role_ids:
            return False
        
        # mentioned_users
        if self.mentioned_users != other.mentioned_users:
            return False
        
        # soundboard_sounds
        if self.soundboard_sounds != other.soundboard_sounds:
            return False
        
        # stickers
        if self.stickers != other.stickers:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        return True
    
    
    def __repr__(self):
        """Returns the message snapshot's representation."""
        repr_parts = ['<', type(self).__name__]
        
        field_added = False
        
        # attachments
        attachments = self.attachments
        if (attachments is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' attachments = ')
            repr_parts.append(repr(attachments))
        
        # components
        components = self.components
        if (components is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' components = ')
            repr_parts.append(repr(components))
        
        # content
        content = self.content
        if (content is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' content = ')
            repr_parts.append(repr(content))
        
        # created_at
        created_at = self.created_at
        if (created_at != DISCORD_EPOCH_START):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' created_at = ')
            repr_parts.append(format(created_at, DATETIME_FORMAT_CODE))
        
        # edited_at
        edited_at = self.edited_at
        if (edited_at is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' edited_at = ')
            repr_parts.append(format(edited_at, DATETIME_FORMAT_CODE))
        
        # embeds
        embeds = self.embeds
        if (embeds is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' embeds = ')
            repr_parts.append(repr(embeds))
        
        # flags
        flags = self.flags
        if flags:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' flags = ')
            repr_parts.append(repr(flags))
        
        # mentioned_role_ids
        mentioned_role_ids = self.mentioned_role_ids
        if mentioned_role_ids:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' mentioned_role_ids = ')
            repr_parts.append(repr(mentioned_role_ids))
        
        # mentioned_users
        mentioned_users = self.mentioned_users
        if mentioned_users:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' mentioned_users = ')
            repr_parts.append(repr(mentioned_users))
        
        # soundboard_sounds
        soundboard_sounds = self.soundboard_sounds
        if (soundboard_sounds is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' soundboard_sounds = ')
            repr_parts.append(repr(soundboard_sounds))
        
        # stickers
        stickers = self.stickers
        if (stickers is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' stickers = ')
            repr_parts.append(repr(stickers))
        
        # type
        message_type = self.type
        if (message_type is not MessageType.default):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' type = ')
            repr_parts.append(message_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(message_type.value))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the message snapshot's hash value."""
        hash_value = 0
        
        # attachments
        attachments = self.attachments
        if (attachments is not None):
            hash_value ^= len(attachments)
            
            for attachment in attachments:
                hash_value ^= hash(attachment)
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= len(components)
            
            for attachment in components:
                hash_value ^= hash(attachment)
        
        # content
        content = self.content
        if (content is not None):
            hash_value ^= hash(content)
        
        # created_at
        created_at = self.created_at
        if (created_at != DISCORD_EPOCH_START):
            hash_value ^= hash(created_at)
        
        # edited_at
        edited_at = self.edited_at
        if (edited_at is not None):
            hash_value ^= hash(edited_at)
        
        # embeds
        embeds = self.embeds
        if (embeds is not None):
            hash_value ^= len(embeds) << 8
            
            for embed in embeds:
                hash_value ^= hash(embed)
        
        # flags
        hash_value ^= self.flags
        
        # mentioned_role_ids
        mentioned_role_ids = self.mentioned_role_ids
        if (mentioned_role_ids is not None):
            hash_value ^= len(mentioned_role_ids) << 21
            
            for role_id in mentioned_role_ids:
                hash_value ^= role_id
        
        # mentioned_users
        mentioned_users = self.mentioned_users
        if (mentioned_users is not None):
            hash_value ^= len(mentioned_users) << 25
            
            for user in mentioned_users:
                hash_value ^= hash(user)
        
        # soundboard_sounds
        soundboard_sounds = self.soundboard_sounds
        if (soundboard_sounds is not None):
            hash_value ^= len(soundboard_sounds) << 27
            
            for soundboard_sound in soundboard_sounds:
                hash_value ^= hash(soundboard_sound)
        
        # stickers
        stickers = self.stickers
        if (stickers is not None):
            hash_value ^= len(stickers)
            
            for attachment in stickers:
                hash_value ^= hash(attachment)
        
        # type
        hash_value ^= hash(self.type)
        
        return hash_value
    
    
    def copy(self):
        """
        Copies the message snapshot.
        
        Returns
        -------
        new : `instance<content<self>>`
        """
        new = object.__new__(type(self))
        new._cache_mentioned_channels = None
        new._state = 0
        
        attachments = self.attachments
        if (attachments is not None):
            attachments = (*attachments,)
        new.attachments = attachments
        
        components = self.components
        if (components is not None):
            components = (*components,)
        new.components = components
        
        new.content = self.content
        new.created_at = self.created_at
        new.edited_at = self.edited_at
        
        embeds = self.embeds
        if (embeds is not None):
            embeds = (*(embed.copy() for embed in embeds),)
        new.embeds = embeds
        
        new.flags = self.flags
        
        mentioned_role_ids = self.mentioned_role_ids
        if (mentioned_role_ids is not None):
            mentioned_role_ids = (*mentioned_role_ids,)
        new.mentioned_role_ids = mentioned_role_ids
        
        mentioned_users = self.mentioned_users
        if (mentioned_users is not None):
            mentioned_users = (*mentioned_users,)
        new.mentioned_users = mentioned_users
        
        soundboard_sounds = self.soundboard_sounds
        if (soundboard_sounds is not None):
            soundboard_sounds = (*soundboard_sounds,)
        new.soundboard_sounds = soundboard_sounds
        
        stickers = self.stickers
        if (stickers is not None):
            stickers = (*stickers,)
        new.stickers = stickers
        
        new.type = self.type
        
        return new
    
    
    def copy_with(
        self,
        *,
        attachments = ...,
        components = ...,
        content = ...,
        created_at = ...,
        edited_at = ...,
        embeds = ...,
        flags = ...,
        mentioned_role_ids = ...,
        mentioned_users = ...,
        message_type = ...,
        soundboard_sounds = ...,
        stickers = ...,
    ):
        """
        Copies the message snapshot with the given fields.
        
        Parameters
        ----------
        attachments : `None | iterable<Attachment>`, Optional (Keyword only)
            The snapshotted message's attachments.
        
        components : `None | iterable<Component>`, Optional (Keyword only)
            The snapshotted message's components.
        
        content : `None | str`, Optional (Keyword only)
            The snapshotted message's content.
        
        created_at : `DateTime`, Optional (Keyword only)
            When the snapshotted message was created.
        
        edited_at : `None | Datetime`, Optional (Keyword only)
            When the snapshotted message was edited.
        
        embeds : `None | iterable<Embed>`, Optional (Keyword only)
            The snapshotted message's embeds.
        
        flags : `MessageFlag | int | None`, Optional (Keyword only)
            The snapshotted message's flags.
        
        mentioned_role_ids : `None | iterable<int>` | iterable<Role>`, Optional (Keyword only)
            The mentioned roles' identifiers.
        
        mentioned_users : `None | iterable<ClientUserBase>`, Optional (Keyword only)
            The mentioned users.
        
        message_type : `MessageType | int | None`, Optional (Keyword only)
            The snapshotted message's type.
        
        soundboard_sounds : `None | iterable<SoundboardSound>`, Optional (Keyword only)
            Soundboard sounds attached to the message.
        
        stickers : `None | iterable<Sticker>`, Optional (Keyword only)
            The snapshotted message's stickers.
        
        Returns
        -------
        new : `instance<content<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's content is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # attachments
        if attachments is ...:
            attachments = self.attachments
            if (attachments is not None):
                attachments = (*attachments,)
        else:
            attachments = validate_attachments(attachments)
        
        # components
        if components is ...:
            components = self.components
            if (components is not None):
                components = (*components,)
        else:
            components = validate_components(components)
        
        # content
        if content is ...:
            content = self.content
        else:
            content = validate_content(content)
        
        # created_at
        if created_at is ...:
            created_at = self.created_at
        else:
            created_at = validate_created_at(created_at)
        
        # edited_at
        if edited_at is ...:
            edited_at = self.edited_at
        else:
            edited_at = validate_edited_at(edited_at)
        
        # embeds
        if embeds is ...:
            embeds = self.embeds
            if (embeds is not None):
                embeds = (*(embed.copy() for embed in embeds),)
        else:
            embeds = validate_embeds(embeds)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # mentioned_role_ids
        if mentioned_role_ids is ...:
            mentioned_role_ids = self.mentioned_role_ids
            if (mentioned_role_ids is not None):
                mentioned_role_ids = (*mentioned_role_ids,)
        else:
            mentioned_role_ids = validate_mentioned_role_ids(mentioned_role_ids)
        
        # mentioned_users
        if mentioned_users is ...:
            mentioned_users = self.mentioned_users
            if (mentioned_users is not None):
                mentioned_users = (*mentioned_users,)
        else:
            mentioned_users = validate_mentioned_users(mentioned_users)
        
        # soundboard_sounds
        if soundboard_sounds is ...:
            soundboard_sounds = self.soundboard_sounds
            if (soundboard_sounds is not None):
                soundboard_sounds = (*soundboard_sounds,)
        else:
            soundboard_sounds = validate_soundboard_sounds(soundboard_sounds)
        
        # stickers
        if stickers is ...:
            stickers = self.stickers
            if (stickers is not None):
                stickers = (*stickers,)
        else:
            stickers = validate_stickers(stickers)
        
        # type
        if message_type is ...:
            message_type = self.type
        else:
            message_type = validate_type(message_type)
        
        # Construct
        new = object.__new__(type(self))
        new._cache_mentioned_channels = None
        new._state = 0
        new.attachments = attachments
        new.components = components
        new.content = content
        new.created_at = created_at
        new.edited_at = edited_at
        new.embeds = embeds
        new.flags = flags
        new.mentioned_role_ids = mentioned_role_ids
        new.mentioned_users = mentioned_users
        new.soundboard_sounds = soundboard_sounds
        new.stickers = stickers
        new.type = message_type
        return new
    
    # Iterators
    
    def iter_attachments(self):
        """
        Iterates over the attachments of the message snapshot.
        
        This method is an iterable generator.
        
        Yields
        ------
        attachment : ``Attachment``
        """
        attachments = self.attachments
        if attachments is not None:
            yield from attachments
    
    
    def iter_components(self):
        """
        Iterates over the components of the message snapshot.
        
        This method is an iterable generator.
        
        Yields
        ------
        attachment : ``Component``
        """
        components = self.components
        if components is not None:
            yield from components
    
    
    def iter_embeds(self):
        """
        Iterates over the embeds of the message snapshot.
        
        This method is an iterable generator.
        
        Yields
        ------
        embed : ``Embed``
        """
        embeds = self.embeds
        if embeds is not None:
            yield from embeds
    
    
    def iter_mentioned_role_ids(self):
        """
        Iterates over the mentioned roles' identifiers in the message.
        
        This method is an iterable generator.
        
        Yields
        ------
        mentioned_role_id : `int`
        """
        mentioned_role_ids = self.mentioned_role_ids
        if (mentioned_role_ids is not None):
            yield from mentioned_role_ids
    
    
    def iter_mentioned_roles(self):
        """
        Iterates over the mentioned roles in the message.
        
        This method is an iterable generator.
        
        Yields
        ------
        mentioned_role : ``Role``
        """
        mentioned_role_ids = self.mentioned_role_ids
        if (mentioned_role_ids is not None):
            for role_id in mentioned_role_ids:
                yield create_partial_role_from_id(role_id)
    
    
    def iter_mentioned_users(self):
        """
        Iterates over the mentioned users in the message.
        
        This method is an iterable generator.
        
        Yields
        ------
        mentioned_user : ``ClientUserBase``
        """
        mentioned_users = self.mentioned_users
        if (mentioned_users is not None):
            yield from mentioned_users
    
    
    def iter_mentioned_channels(self):
        """
        Iterates over the mentioned channels in the message.
        
        This method is an iterable generator.
        
        Yields
        ------
        mentioned_channel : ``Channel``
        """
        mentioned_channels = self.mentioned_channels
        if (mentioned_channels is not None):
            yield from mentioned_channels
    
    
    def iter_soundboard_sounds(self):
        """
        Iterates over the soundboard sounds of the message.
        
        This method is an iterable generator.
        
        Yields
        ------
        soundboard_sounds : ``SoundboardSound``
        """
        soundboard_sounds = self.soundboard_sounds
        if (soundboard_sounds is not None):
            yield from soundboard_sounds
    
    
    def iter_stickers(self):
        """
        Iterates over the stickers of the message snapshot.
        
        This method is an iterable generator.
        
        Yields
        ------
        attachment : ``Sticker``
        """
        stickers = self.stickers
        if stickers is not None:
            yield from stickers
    
    
    # get many
    
    
    @property
    def mentioned_roles(self):
        """
        The mentioned roles by the message. If there is non, returns `None`.
        
        Returns
        -------
        role_mentions : `None | tuple<Role>`
        """
        mentioned_role_ids = self.mentioned_role_ids
        if (mentioned_role_ids is not None):
            return (*(create_partial_role_from_id(role_id) for role_id in mentioned_role_ids),)
    
    
    
    def _get_mentioned_channels(self):
        """
        Looks up the ``.contents`` of the message and searches channel mentions in them.
        
        Invalid channel mentions are ignored.
        
        Returns
        -------
        channel_mentions : `None`, `tuple` of ``Channel``
            The parsed channel mentions.
        """
        mentioned_channels = None
        
        content = self.content
        if content is None:
            return mentioned_channels
        
        for channel_id in CHANNEL_MENTION_RP.findall(content):
            channel_id = int(channel_id)
            try:
                channel = CHANNELS[channel_id]
            except KeyError:
                continue
            
            if mentioned_channels is None:
                mentioned_channels = set()
            
            mentioned_channels.add(channel)
        
        if mentioned_channels is not None:
            mentioned_channels = tuple(sorted(mentioned_channels, key = id_sort_key))
        
        return mentioned_channels
    
    
    @property
    def mentioned_channels(self):
        """
        The mentioned channels by the message. If there is non returns `None`.
        
        Returns
        -------
        mentioned_channels : `None`, `tuple` of ``Channel``
        """
        state = self._state
        if state & MESSAGE_STATE_MASK_CACHE_MENTIONED_CHANNELS:
            mentioned_channels = self._cache_mentioned_channels
        else:
            mentioned_channels = self._get_mentioned_channels()
            self._cache_mentioned_channels = mentioned_channels
            self._state = state | MESSAGE_STATE_MASK_CACHE_MENTIONED_CHANNELS
        
        return mentioned_channels
    
    
    # get one
    
    @property
    def attachment(self):
        """
        Returns the first attachment in the message snapshot.

        Returns
        -------
        attachment : `None | Attachment`
        """
        attachments = self.attachments
        if attachments is not None:
            return attachments[0]
    
    
    @property
    def embed(self):
        """
        Returns the first embed in the message snapshot.

        Returns
        -------
        embed : `None | Embed`
        """
        embeds = self.embeds
        if embeds is not None:
            return embeds[0]
    
    
    @property
    def soundboard_sound(self):
        """
        Returns the message's first soundboard sound.
        
        Returns
        -------
        soundboard_sound : `None | SoundboardSound`
        """
        soundboard_sounds = self.soundboard_sounds
        if (soundboard_sounds is not None):
            return soundboard_sounds[0]
    
    
    @property
    def sticker(self):
        """
        Returns the first sticker in the message snapshot.

        Returns
        -------
        Sticker : `None | Sticker`
        """
        stickers = self.stickers
        if stickers is not None:
            return stickers[0]



__all__ = ('TeamMember',)

from scarletio import RichAttributeErrorBaseType

from ...user import ZEROUSER

from .fields import (
    parse_role, parse_state, parse_user, put_role, put_state, put_user,
    validate_role, validate_state, validate_user
)
from .preinstanced import TeamMemberRole, TeamMembershipState


class TeamMember(RichAttributeErrorBaseType):
    """
    Represents a member of a ``Team``.
    
    Attributes
    ----------
    role : `None`, `tuple` of ``TeamMemberRole``
        The role of the team member.
    state : ``TeamMembershipState``
        The state of the team member. A member can be invited or can have the invite already accepted.
    user : ``ClientUserBase``
        The corresponding user account of the team member object.
    """
    __slots__ = ('role', 'state', 'user',)
    
    def __new__(cls, *, role = ..., state = ..., user = ...):
        """
        Creates a new team member from the given parameters.
        
        Parameters
        ----------
        role : `None`, `str`, ``TeamMemberRole``, Optional (Keyword only)
            The role of the team member.
        
        state : `int`, ``TeamMembershipState``, Optional (Keyword only)
            The state of the team member.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The corresponding user account of the team member.
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # role
        if role is ...:
            role = TeamMemberRole.none
        else:
            role = validate_role(role)
        
        # state
        if state is ...:
            state = TeamMembershipState.none
        else:
            state = validate_state(state)
        
        # user
        if user is ...:
            user = ZEROUSER
        else:
            user = validate_user(user)
        
        self = object.__new__(cls)
        self.role = role
        self.state = state
        self.user = user
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a `TeamMember` object from the data sent by Discord.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Team member data received from Discord.
        """
        self = object.__new__(cls)
        self.role = parse_role(data)
        self.state = parse_state(data)
        self.user = parse_user(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the team member to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_role(self.role, data, defaults)
        put_state(self.state, data, defaults)
        
        if include_internals:
            put_user(self.user, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the team member's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # user
        repr_parts.append(' user = ')
        repr_parts.append(repr(self.user))
        
        # state
        repr_parts.append(', state = ')
        repr_parts.append(self.state.name)
        
        # role
        repr_parts.append(', role = ')
        repr_parts.append(self.role.name)
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the team member's hash value, what is equal to it's user's id."""
        hash_value = 0
        
        # role
        hash_value ^= hash(self.role.value)
        
        # state
        hash_value ^= self.state.value
        
        # user
        hash_value ^= hash(self.user)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two team members are equal."""
        if type(self) is not type(other):
            return False
        
        # role
        if (self.role is not other.role):
            return False
        
        # state
        if (self.state is not other.state):
            return False
        
        # user
        if (self.user != other.user):
            return False
        
        return True
    
    
    def __gt__(self, other):
        """Returns whether self is less than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.user > other.user
    
    
    def __lt__(self, other):
        """Returns whether self is greater than other."""
        if type(self) is not type(other):
            return NotImplemented
        
        return self.user < other.user
    
    
    def copy(self):
        """
        Copies the team member.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.role = self.role
        new.state = self.state
        new.user = self.user
        return new
    
    
    def copy_with(self, *, role = ..., state = ..., user = ...):
        """
        Copies the team member with the the defined fields.
        
        Parameters
        ----------
        role : `None`, `str`, ``TeamMemberRole``, Optional (Keyword only)
            The role of the team member.
        
        state : `int`, ``TeamMembershipState``, Optional (Keyword only)
            The state of the team member.
        
        user : ``ClientUserBase``, Optional (Keyword only)
            The corresponding user account of the team member.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # role
        if role is ...:
            role = self.role
        else:
            role = validate_role(role)
        
        # state
        if state is ...:
            state = self.state
        else:
            state = validate_state(state)
        
        # user
        if user is ...:
            user = self.user
        else:
            user = validate_user(user)
        
        new = object.__new__(type(self))
        new.role = role
        new.state = state
        new.user = user
        return new



"""
Before loading hata, it checks for related environmental variables, which are:

HATA_ALLOW_DEBUG_MESSAGES : `bool` = `False`
    Whether debug messages should be show (experimental).

HATA_API_ENDPOINT : `None`, `str` = `None`
    The api endpoint to use instead of the Discord's default.

HATA_API_VERSION : `int` = `10`
    The Discord api version used by hata. The accepted values are `6`, `7`, `8`, `9` and `10`.
    
    If given as any other value, a warning message will show up. Tho, if given `6` a deprecation warning will be still
    present.

HATA_CACHE_PRESENCE : `bool` = `True`
    Whether hata should enable presence related attributes and dispatch users presence related events. By disabling it,
    ``User.status``, ``User.statuses``, ``User.platform``, ``User.activities``, ``User.activity`` will be disabled. And
    each will be replaced by a dummy property.
    
    If `HATA_CACHE_USERS` is defined as `False`, `HATA_CACHE_PRESENCE` will be set as `False` as well.

HATA_CACHE_USERS : `bool` = `True`
    Whether hata should cache users. Disabling it can cause many hata features to disappear.

HATA_CDN_ENDPOINT : `None`, `str` = `None`
    The cdn (content delivery network) endpoint to use instead of the Discord's default.

HATA_DISCORD_ENDPOINT : `None`, `str` = `None`
    The endpoint of Discord, to use instead of it's own.

HATA_DOCS_ENABLED : `bool` = `True`
    Whether hata should be loaded with docstrings.
    
    > Experimental, not the full wrappers supports it yet.
    
    If python is run with `-OO`, then this always defaults to `False`.

HATA_LIBRARY_AGENT_APPENDIX : `str` = `None`
    Library agent appendix used instead of the default generated one.

HATA_LIBRARY_NAME : `str` = `'hata'`
    Library name used when identifying.

HATA_LIBRARY_URL : `str` = `'https://github.com/HuyaneMatsu/hata'`
    Library url used at user agents.

HATA_LIBRARY_VERSION : `str` = `None`
    Library version used in user agents.

HATA_MESSAGE_CACHE_SIZE : `int` = `10`
    The default message cache size per channel.

HATA_RICH_DISCORD_EXCEPTION : `bool` = `False`
    Whether ``DiscordException``-s should show the request data as well.

HATA_STATUS_ENDPOINT : `None`, `str` = `None`
    Discord status endpoint.
"""

from .env_getter import *
from .getters import *
from .loading import *
from .parsing import *
from .variables import *


__all__ = (
    *env_getter.__all__,
    *getters.__all__,
    *loading.__all__,
    *parsing.__all__,
    *variables.__all__,
)



__all__ = ('ActivityMetadataRich',)

from scarletio import copy_docs

from .base import ActivityMetadataBase
from .fields import (
    parse_application_id, parse_assets, parse_buttons, parse_created_at, parse_details, parse_flags, parse_id,
    parse_name, parse_party, parse_secrets, parse_session_id, parse_state, parse_sync_id, parse_timestamps, parse_url,
    put_application_id, put_assets, put_buttons, put_created_at, put_details, put_flags,
    put_id, put_name, put_party, put_secrets, put_session_id, put_state, put_sync_id,
    put_timestamps, put_url, validate_application_id, validate_assets, validate_buttons, validate_created_at,
    validate_details, validate_flags, validate_id, validate_name, validate_party, validate_secrets, validate_session_id,
    validate_state, validate_sync_id, validate_timestamps, validate_url
)
from .flags import ActivityFlag


class ActivityMetadataRich(ActivityMetadataBase):
    """
    Represents a Discord rich activity.
    
    Attributes
    ----------
    application_id : `int`
        The id of the activity's application. Defaults to `0`.
    
    assets : `None | ActivityAssets`
        The activity's assets. Defaults to `None`.
    
    buttons : `None | tuple<str>`
        The labels of the buttons on the activity.
    
    created_at : `None | DateTime`
        When the activity was created. Defaults to Discord epoch.
    
    details : `None | str`
        What the player is currently doing. Defaults to `None`.
    
    flags : ``ActivityFlag``
        The flags of the activity. Defaults to `ActivityFlag(0)`
    
    id : `int`
        The id of the activity. Defaults to `0`.
    
    name : `str`
        The activity's name.
    
    party : `None | ActivityParty`
        The activity's party. Defaults to `None`.
    
    secrets : `None | ActivitySecrets`
        The activity's secrets. Defaults to `None`.
    
    session_id : `None | str`
        Spotify activity's session's id. Defaults to `None`.
    
    state : `None | str`
        The player's current party status. Defaults to `None`.
    
    sync_id : `None | str`
        The id of the currently playing track of a spotify activity. Defaults to `None`.
    
    timestamps : `None | ActivityTimestamps`
        The activity's timestamps.
    
    url : `None | str`
        The url of the stream (Twitch or Youtube only). Defaults to `None`.
    """
    __slots__ = (
        'application_id', 'assets', 'buttons', 'created_at', 'details', 'flags', 'id', 'name', 'party', 'secrets',
        'session_id', 'state', 'sync_id', 'timestamps', 'url'
    )
    
    
    def __new__(
        cls,
        *,
        activity_id = ...,
        application_id = ...,
        assets = ...,
        buttons = ...,
        created_at = ...,
        details = ...,
        flags = ...,
        name = ...,
        party = ...,
        secrets = ...,
        session_id = ...,
        state = ...,
        sync_id = ...,
        timestamps = ...,
        url = ...,
    ):
        """
        Creates a new rich activity metadata from the given parameters.
        
        Attributes
        ----------
        activity_id : `int`, Optional (Keyword only)
            The id of the activity.
        
        application_id : `int`, Optional (Keyword only)
            The id of the activity's application.
        
        assets : `None | ActivityAssets`, Optional (Keyword only)
            The activity's assets
    
        buttons : `None | str | iterable<str>`, Optional (Keyword only)
            The labels of the buttons on the activity.
        
        created_at : `None | Datetime`, Optional (Keyword only)
            When the activity was created.
        
        details : `None | str`, Optional (Keyword only)
            What the player is currently doing.
        
        flags : `ActivityFlag | int`, Optional (Keyword only)
            The flags of the activity.
        
        name : `str`, Optional (Keyword only)
            The activity's name.
        
        party : `None | ActivityParty`, Optional (Keyword only)
            The activity's party.
        
        secrets : `None | ActivitySecrets`, Optional (Keyword only)
            The activity's secrets.
        
        session_id : `None | str`, Optional (Keyword only)
            Spotify activity's session's id.
        
        state : `None | str`, Optional (Keyword only)
            The player's current party status.
        
        sync_id : `None | str`, Optional (Keyword only)
            The id of the currently playing track of a spotify activity.
        
        timestamps : `None | ActivityTimestamps`, Optional (Keyword only)
            The activity's timestamps.
        
        url : `None | str`, Optional (Keyword only)
            The url of the stream (Twitch or Youtube only).
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        # application_id
        if application_id is ...:
            application_id = 0
        else:
            application_id = validate_application_id(application_id)
        
        # assets
        if assets is ...:
            assets = None
        else:
            assets = validate_assets(assets)
        
        if buttons is ...:
            buttons = None
        else:
            buttons = validate_buttons(buttons)
        
        # created_at
        if created_at is ...:
            created_at = None
        else:
            created_at = validate_created_at(created_at)
        
        # details
        if details is ...:
            details = None
        else:
            details = validate_details(details)
        
        # flags
        if flags is ...:
            flags = ActivityFlag()
        else:
            flags = validate_flags(flags)
        
        # activity_id
        if activity_id is ...:
            activity_id = 0
        else:
            activity_id = validate_id(activity_id)
        
        # name
        if name is ...:
            name = ''
        else:
            name = validate_name(name)
        
        # party
        if party is ...:
            party = None
        else:
            party = validate_party(party)
        
        # secrets
        if secrets is ...:
            secrets = None
        else:
            secrets = validate_secrets(secrets)
        
        # session_id
        if session_id is ...:
            session_id = None
        else:
            session_id = validate_session_id(session_id)
        
        # state
        if state is ...:
            state = None
        else:
            state = validate_state(state)
        
        # sync_id
        if sync_id is ...:
            sync_id = None
        else:
            sync_id = validate_sync_id(sync_id)
        
        # timestamps
        if timestamps is ...:
            timestamps = None
        else:
            timestamps = validate_timestamps(timestamps)
        
        # url
        if url is ...:
            url = None
        else:
            url = validate_url(url)
        
        # Construct
        self = object.__new__(cls)
        self.application_id = application_id
        self.assets = assets
        self.created_at = created_at
        self.buttons = buttons
        self.details = details
        self.flags = flags
        self.id = activity_id
        self.name = name
        self.party = party
        self.secrets = secrets
        self.session_id = session_id
        self.state = state
        self.sync_id = sync_id
        self.timestamps = timestamps
        self.url = url
        return self
    
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            activity_id = keyword_parameters.pop('activity_id', ...),
            application_id = keyword_parameters.pop('application_id', ...),
            assets = keyword_parameters.pop('assets', ...),
            buttons = keyword_parameters.pop('buttons', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            details = keyword_parameters.pop('details', ...),
            flags = keyword_parameters.pop('flags', ...),
            name = keyword_parameters.pop('name', ...),
            party = keyword_parameters.pop('party', ...),
            secrets = keyword_parameters.pop('secrets', ...),
            session_id = keyword_parameters.pop('session_id', ...),
            state = keyword_parameters.pop('state', ...),
            sync_id = keyword_parameters.pop('sync_id', ...),
            timestamps = keyword_parameters.pop('timestamps', ...),
            url = keyword_parameters.pop('url', ...),
        )
    
    
    @copy_docs(ActivityMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', self.__class__.__name__]
        
        repr_parts.append(' name = ')
        repr_parts.append(repr(self.name))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ActivityMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # application_id
        hash_value ^= self.application_id
        
        # assets
        assets = self.assets
        if (assets is not None):
            hash_value ^= hash(assets)
        
        # buttons
        buttons = self.buttons
        if (buttons is not None):
            hash_value ^= len(buttons) << 2
            for button in buttons:
                hash_value ^= hash(button)
        
        # created_at
        created_at = self.created_at
        if (created_at is not None):
            hash_value ^= hash(created_at)
        
        # details
        details = self.details
        if (details is not None):
            hash_value ^= hash(details)
        
        # flags
        hash_value ^= self.flags
        
        # id
        hash_value ^= self.id
        
        # name
        hash_value ^= hash(self.name)
        
        # party
        party = self.party
        if (party is not None):
            hash_value ^= hash(party)
        
        # secrets
        secrets = self.secrets
        if (secrets is not None):
            hash_value ^= hash(secrets)
        
        # session_id
        session_id = self.session_id
        if (session_id is not None):
            hash_value ^= hash(session_id)
        
        # state
        state = self.state
        if (state is not None):
            hash_value ^= hash(state)
        
        # sync_id
        sync_id = self.sync_id
        if (sync_id is not None):
            hash_value ^= hash(sync_id)
        
        # timestamps
        timestamps = self.timestamps
        if (timestamps is not None):
            hash_value ^= hash(timestamps)
        
        # url
        url = self.url
        if (url is not None):
            hash_value ^= hash(url)
        
        return hash_value
    
    
    @copy_docs(ActivityMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # application_id
        if self.application_id != other.application_id:
            return False
        
        # assets
        if self.assets != other.assets:
            return False
        
        # buttons
        if self.buttons != other.buttons:
            return False
        
        # created_at
        if self.created_at != other.created_at:
            return False
        
        # details
        if self.details != other.details:
            return False
        
        # flags
        if self.flags != other.flags:
            return False
        
        # id
        if self.id != other.id:
            return False
        
        # name
        if self.name != other.name:
            return False
        
        # party
        if self.party != other.party:
            return False
        
        # secrets
        if self.secrets != other.secrets:
            return False
        
        # session_id
        if self.session_id != other.session_id:
            return False
        
        # state
        if self.state != other.state:
            return False
        
        # sync_id
        if self.sync_id != other.sync_id:
            return False
        
        # timestamps
        if self.timestamps != other.timestamps:
            return False
        
        # url
        if self.url != other.url:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ActivityMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.application_id = parse_application_id(data)
        self.id = parse_id(data)
        self._update_attributes(data)
        return self
    
    
    @copy_docs(ActivityMetadataBase.to_data)
    def to_data(self, *, defaults = False, include_internals = False, user = False):
        data = {}
        
        put_name(self.name, data, defaults)
        put_state(self.state, data, defaults)
        put_url(self.url, data, defaults)
        
        if user or include_internals:
            put_assets(self.assets, data, defaults)
            put_buttons(self.buttons, data, defaults)
            put_details(self.details, data, defaults)
            put_party(self.party, data, defaults)
            put_secrets(self.secrets, data, defaults)
            put_timestamps(self.timestamps, data, defaults)
        
        if include_internals:
            # receive only?
            put_application_id(self.application_id, data, defaults)
            put_created_at(self.created_at, data, defaults)
            put_id(self.id, data, defaults)
            
            # spotify only?
            put_flags(self.flags, data, defaults)
            put_session_id(self.session_id, data, defaults)
            put_sync_id(self.sync_id, data, defaults)
        
        return data
    
    
    @copy_docs(ActivityMetadataBase._update_attributes)
    def _update_attributes(self, data):
        # application_id & id never changes
        
        self.assets = parse_assets(data)
        self.created_at = parse_created_at(data)
        self.buttons = parse_buttons(data)
        self.details = parse_details(data)
        self.flags = parse_flags(data)
        self.name = parse_name(data)
        self.party = parse_party(data)
        self.secrets = parse_secrets(data)
        self.session_id = parse_session_id(data)
        self.state = parse_state(data)
        self.sync_id = parse_sync_id(data)
        self.timestamps = parse_timestamps(data)
        self.url = parse_url(data)
    
    
    @copy_docs(ActivityMetadataBase._difference_update_attributes)
    def _difference_update_attributes(self, data):
        old_attributes = {}
        
        # application_id & id never changes
        
        # assets
        assets = parse_assets(data)
        if self.assets != assets:
            old_attributes['assets'] = self.assets
            self.assets = assets
        
        # buttons
        buttons = parse_buttons(data)
        if self.buttons != buttons:
            old_attributes['buttons'] = self.buttons
            self.buttons = buttons
        
        # created_at
        created_at = parse_created_at(data)
        if self.created_at != created_at:
            old_attributes['created_at'] = self.created_at
            self.created_at = created_at
        
        # details
        details = parse_details(data)
        if self.details != details:
            old_attributes['details'] = self.details
            self.details = details
        
        # flags
        flags = parse_flags(data)
        if self.flags != flags:
            old_attributes['flags'] = self.flags
            self.flags = flags
        
        # name
        name = parse_name(data)
        if self.name != name:
            old_attributes['name'] = self.name
            self.name = name
        
        # party
        party = parse_party(data)
        if self.party != party:
            old_attributes['party'] = self.party
            self.party = party
        
        # secrets
        secrets = parse_secrets(data)
        if self.secrets != secrets:
            old_attributes['secrets'] = self.secrets
            self.secrets = secrets
        
        # session_id
        session_id = parse_session_id(data)
        if self.session_id != session_id:
            old_attributes['session_id'] = self.session_id
            self.session_id = session_id
        
        # state
        state = parse_state(data)
        if self.state != state:
            old_attributes['state'] = self.state
            self.state = state
        
        # sync_id
        sync_id = parse_sync_id(data)
        if self.sync_id != sync_id:
            old_attributes['sync_id'] = self.sync_id
            self.sync_id = sync_id
        
        # timestamps
        timestamps = parse_timestamps(data)
        if self.timestamps != timestamps:
            old_attributes['timestamps'] = self.timestamps
            self.timestamps = timestamps
        
        # url
        url = parse_url(data)
        if self.url != url:
            old_attributes['url'] = self.url
            self.url = url
        
        return old_attributes
    
    
    @copy_docs(ActivityMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.application_id = self.application_id
        assets = self.assets
        if (assets is not None):
            assets = assets.copy()
        new.assets = assets
        
        buttons = self.buttons
        if (buttons is not None):
            buttons = (*buttons,)
        new.buttons = buttons
        
        new.created_at = self.created_at
        new.details = self.details
        new.flags = self.flags
        new.id = self.id
        new.name = self.name
        party = self.party
        if (party is not None):
            party = party.copy()
        new.party = party
        secrets = self.secrets
        if (secrets is not None):
            secrets = secrets.copy()
        new.secrets = secrets
        new.session_id = self.session_id
        new.state = self.state
        new.sync_id = self.sync_id
        timestamps = self.timestamps
        if (timestamps is not None):
            timestamps = timestamps.copy()
        new.timestamps = timestamps
        new.url = self.url
        return new
    
    
    def copy_with(
        self, 
        *,
        activity_id = ...,
        application_id = ...,
        assets = ...,
        buttons = ...,
        created_at = ...,
        details = ...,
        flags = ...,
        name = ...,
        party = ...,
        secrets = ...,
        session_id = ...,
        state = ...,
        sync_id = ...,
        timestamps = ...,
        url = ...,
    ):
        """
        Copies the rich activity metadata with the given fields.
        
        Attributes
        ----------
        activity_id : `int`, Optional (Keyword only)
            The id of the activity.
        
        application_id : `int`, Optional (Keyword only)
            The id of the activity's application.
        
        assets : `None | ActivityAssets`, Optional (Keyword only)
            The activity's assets
        
        buttons : `None | str | iterable<str>`, Optional (Keyword only)
            The labels of the buttons on the activity.
        
        created_at : `None | Datetime`, Optional (Keyword only)
            When the activity was created.
        
        details : `None | str`, Optional (Keyword only)
            What the player is currently doing.
        
        flags : `ActivityFlag | int`, Optional (Keyword only)
            The flags of the activity.
        
        name : `str`, Optional (Keyword only)
            The activity's name.
        
        party : `None | ActivityParty`, Optional (Keyword only)
            The activity's party.
        
        secrets : `None | ActivitySecrets`, Optional (Keyword only)
            The activity's secrets.
        
        session_id : `None | str`, Optional (Keyword only)
            Spotify activity's session's id.
        
        state : `None | str`, Optional (Keyword only)
            The player's current party status.
        
        sync_id : `None | str`, Optional (Keyword only)
            The id of the currently playing track of a spotify activity.
        
        timestamps : `None | ActivityTimestamps`, Optional (Keyword only)
            The activity's timestamps.
        
        url : `None | str`, Optional (Keyword only)
            The url of the stream (Twitch or Youtube only).
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        # application_id
        if application_id is ...:
            application_id = self.application_id
        else:
            application_id = validate_application_id(application_id)
        
        # assets
        if assets is ...:
            assets = self.assets
            if (assets is not None):
                assets = assets.copy()
        else:
            assets = validate_assets(assets)
        
        # buttons
        if buttons is ...:
            buttons = self.buttons
            if (buttons is not None):
                buttons = (*buttons,)
        else:
            buttons = validate_buttons(buttons)
        
        # created_at
        if created_at is ...:
            created_at = self.created_at
        else:
            created_at = validate_created_at(created_at)
        
        # details
        if details is ...:
            details = self.details
        else:
            details = validate_details(details)
        
        # flags
        if flags is ...:
            flags = self.flags
        else:
            flags = validate_flags(flags)
        
        # activity_id
        if activity_id is ...:
            activity_id = self.id
        else:
            activity_id = validate_id(activity_id)
        
        # name
        if name is ...:
            name = self.name
        else:
            name = validate_name(name)
        
        # party
        if party is ...:
            party = self.party
            if (party is not None):
                party = party.copy()
        else:
            party = validate_party(party)
        
        # secrets
        if secrets is ...:
            secrets = self.secrets
            if (secrets is not None):
                secrets = secrets.copy()
        else:
            secrets = validate_secrets(secrets)
        
        # session_id
        if session_id is ...:
            session_id = self.session_id
        else:
            session_id = validate_session_id(session_id)
        
        # state
        if state is ...:
            state = self.state
        else:
            state = validate_state(state)
        
        # sync_id
        if sync_id is ...:
            sync_id = self.sync_id
        else:
            sync_id = validate_sync_id(sync_id)
        
        # timestamps
        if timestamps is ...:
            timestamps = self.timestamps
            if (timestamps is not None):
                timestamps = timestamps.copy()
        else:
            timestamps = validate_timestamps(timestamps)
        
        # url
        if url is ...:
            url = self.url
        else:
            url = validate_url(url)
        
        # Construct
        new = object.__new__(type(self))
        new.application_id = application_id
        new.assets = assets
        new.buttons = buttons
        new.created_at = created_at
        new.details = details
        new.flags = flags
        new.id = activity_id
        new.name = name
        new.party = party
        new.secrets = secrets
        new.session_id = session_id
        new.state = state
        new.sync_id = sync_id
        new.timestamps = timestamps
        new.url = url
        return new
    
    
    @copy_docs(ActivityMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            activity_id = keyword_parameters.pop('activity_id', ...),
            application_id = keyword_parameters.pop('application_id', ...),
            assets = keyword_parameters.pop('assets', ...),
            buttons = keyword_parameters.pop('buttons', ...),
            created_at = keyword_parameters.pop('created_at', ...),
            details = keyword_parameters.pop('details', ...),
            flags = keyword_parameters.pop('flags', ...),
            name = keyword_parameters.pop('name', ...),
            party = keyword_parameters.pop('party', ...),
            secrets = keyword_parameters.pop('secrets', ...),
            session_id = keyword_parameters.pop('session_id', ...),
            state = keyword_parameters.pop('state', ...),
            sync_id = keyword_parameters.pop('sync_id', ...),
            timestamps = keyword_parameters.pop('timestamps', ...),
            url = keyword_parameters.pop('url', ...),
        )



from .base import *
from .client_base import *
from .client_shard import *
from .client_sharder import *
from .constants import *
from .heartbeat import *
from .rate_limit import *
from .utils import *
from .voice import *
from .voice_base import *


__all__ = (
    *base.__all__,
    *client_base.__all__,
    *client_shard.__all__,
    *client_sharder.__all__,
    *constants.__all__,
    *heartbeat.__all__,
    *rate_limit.__all__,
    *utils.__all__,
    *voice.__all__,
    *voice_base.__all__,
)



from .attachment import *
from .constants import *
from .fields import *
from .flags import *


__all__ = (
    *attachment.__all__,
    *constants.__all__,
    *fields.__all__,
    *flags.__all__,
)



__all__ = ('InteractionMetadataMessageComponent',)

from scarletio import copy_docs

from ...component import Component, ComponentType

from .base import InteractionMetadataBase
from .fields import (
    parse_component_type, parse_custom_id, parse_resolved, parse_values, put_component_type, put_custom_id,
    put_resolved, put_values, validate_component_type, validate_custom_id, validate_resolved, validate_values
)


class InteractionMetadataMessageComponent(InteractionMetadataBase):
    """
    Interaction metadata used when the interaction was triggered by an application command's auto completion.
    
    Parameters
    ----------
    component_type : ``ComponentType``
        The used component's type.
    
    custom_id : `None`, `str`
        Component or form interaction's custom identifier.
    
    resolved : `None`, ``Resolved``
        Contains the received entities.
    
    values : `None`, `tuple` of `str`
        Values selected by the user.
    """
    __slots__ = ('component_type', 'custom_id', 'resolved', 'values')
    
    def __new__(
        cls,
        *,
        component_type = ...,
        custom_id = ...,
        resolved = ...,
        values = ...,
    ):
        """
        Creates a new interaction metadata from the given parameters.
        
        Parameters
        ----------
        component_type : ``ComponentType``, Optional (Keyword only)
            The used component's type.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Component or form interaction's custom identifier.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        values : `None`, `tuple` of `str`, Optional (Keyword only)
            Values selected by the user. Applicable for component interactions.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = ComponentType.none
        else:
            component_type = validate_component_type(component_type)
        
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # resolved
        if resolved is ...:
            resolved = None
        else:
            resolved = validate_resolved(resolved)
        
        # values
        if values is ...:
            values = None
        else:
            values = validate_values(values)
        
        # Construct
        self = object.__new__(cls)
        self.component_type = component_type
        self.custom_id = custom_id
        self.resolved = resolved
        self.values = values
        return self
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase._create_empty)
    def _create_empty(cls):
        self = object.__new__(cls)
        self.component_type = ComponentType.none
        self.custom_id = None
        self.resolved = None
        self.values = None
        return self
    
    
    @copy_docs(InteractionMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        new.component_type = self.component_type
        new.custom_id = self.custom_id
        
        resolved = self.resolved
        if (resolved is not None):
            resolved = resolved.copy()
        new.resolved = resolved
        
        values = self.values
        if (values is not None):
            values = (*values,)
        new.values = values
        
        return new
    
    def copy_with(
        self,
        *,
        component_type = ...,
        custom_id = ...,
        resolved = ...,
        values = ...,
    ):
        """
        Copies the interaction metadata with the given fields.
        
        Parameters
        ----------
        component_type : ``ComponentType``, Optional (Keyword only)
            The used component's type.
        
        custom_id : `None`, `str`, Optional (Keyword only)
            Component or form interaction's custom identifier.
        
        resolved : `None`, ``Resolved``, Optional (Keyword only)
            Contains the received entities.
        
        values : `None`, `tuple` of `str`, Optional (Keyword only)
            Values selected by the user. Applicable for component interactions.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = self.component_type
        else:
            component_type = validate_component_type(component_type)
        
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # resolved
        if resolved is ...:
            resolved = self.resolved
            if (resolved is not None):
                resolved = resolved.copy()
        else:
            resolved = validate_resolved(resolved)
        
        # values
        if values is ...:
            values = self.values
            if (values is not None):
                values = (*values,)
        else:
            values = validate_values(values)
        
        # Construct
        new = object.__new__(type(self))
        new.component_type = component_type
        new.custom_id = custom_id
        new.resolved = resolved
        new.values = values
        return new
    
    
    @classmethod
    @copy_docs(InteractionMetadataBase.from_data)
    def from_data(cls, data, guild_id = 0):
        self = object.__new__(cls)
        self.component_type = parse_component_type(data)
        self.custom_id = parse_custom_id(data)
        self.resolved = parse_resolved(data, guild_id)
        self.values = parse_values(data)
        return self
    
    
    @copy_docs(InteractionMetadataBase.to_data)
    def to_data(self, *, defaults = False, guild_id = 0):
        data = {}
        put_component_type(self.component_type, data, defaults)
        put_custom_id(self.custom_id, data, defaults)
        put_resolved(self.resolved, data, defaults, guild_id = guild_id)
        put_values(self.values, data, defaults)
        return data
    
    
    @copy_docs(InteractionMetadataBase._put_attribute_representations_into)
    def _put_attribute_representations_into(self, repr_parts):
        # component_type
        repr_parts.append(' component_type = ')
        repr_parts.append(repr(self.component_type))
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            repr_parts.append(', custom_id = ')
            repr_parts.append(repr(custom_id))
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            repr_parts.append(', resolved = ')
            repr_parts.append(repr(resolved))
        
        # values
        values = self.values
        if (values is not None):
            repr_parts.append(', values = ')
            repr_parts.append(repr(values))
        
        return True
    
    
    @copy_docs(InteractionMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # component_type
        hash_value ^= self.component_type.value
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # resolved
        resolved = self.resolved
        if (resolved is not None):
            hash_value ^= hash(resolved)
        
        # values
        values = self.values
        if (values is not None):
            hash_value ^= len(values)
            
            for option in values:
                hash_value ^= hash(option)
        
        return hash_value
    
    
    @copy_docs(InteractionMetadataBase.__eq__)
    def __eq__(self, other):
        other_type = type(other)
        if other_type is type(self):
            return self._is_equal_same_type(other)
        
        # Compare with components.
        if issubclass(other_type, Component):
            # Check `type` before `custom_id`
            
            # type
            if self.component_type is not other.type:
                return False
            
            # custom_id
            if self.custom_id != other.custom_id:
                return False
            
            return True
        
        return NotImplemented
    
    
    @copy_docs(InteractionMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # component_type
        if self.component_type is not other.component_type:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # resolved
        if self.resolved != other.resolved:
            return False
        
        # values
        if self.values != other.values:
            return False
        
        return True



from .aead_aes256_gcm_rtpsize import *
from .aead_xchacha20_poly1305_rtpsize import *
from .base import *
from .xsalsa20_poly1305 import *


__all__ = (
    *aead_aes256_gcm_rtpsize.__all__,
    *aead_xchacha20_poly1305_rtpsize.__all__,
    *base.__all__,
    *xsalsa20_poly1305.__all__,
)

# Construct `AVAILABLE_ENCRYPTION_ADAPTERS`

from .aead_aes256_gcm_rtpsize import EncryptionAdapter__aead_aes256_gcm_rtpsize
from .aead_xchacha20_poly1305_rtpsize import EncryptionAdapter__aead_xchacha20_poly1305_rtpsize
from .xsalsa20_poly1305 import EncryptionAdapter__xsalsa20_poly1305


AVAILABLE_ENCRYPTION_ADAPTERS = (*sorted(
    (
        encryption_adapter for encryption_adapter in
        (
            EncryptionAdapter__aead_aes256_gcm_rtpsize, # -> not working actually, lmeow
            EncryptionAdapter__aead_xchacha20_poly1305_rtpsize, # -> not working actually, lmeow
            EncryptionAdapter__xsalsa20_poly1305,
        )
        if encryption_adapter.available
    ),
    key = (lambda encryption_adapter : encryption_adapter.priority),
    reverse = True,
),)


# Import `EncryptionAdapterBase` to shorten imports

from .base import EncryptionAdapterBase



from .audit_log_entry_detail_conversion import *
from .audit_log_entry_detail_conversion_group import *


__all__ = (
    *audit_log_entry_detail_conversion.__all__,
    *audit_log_entry_detail_conversion_group.__all__,
)



__all__ = ('VerificationScreenStepType',)

from scarletio import copy_docs

from ...bases import Preinstance as P, PreinstancedBase


class VerificationScreenStepType(PreinstancedBase, value_type = str):
    """
    Represents a type of a ``VerificationScreenStep``.

    Attributes
    ----------
    name : `str`
        The verification screen step type's name.
    
    value : `str`
        The Discord side identifier value of the verification step types.
    
    Type Attributes
    ---------------
    Every predefined verification screen step type can be accessed as type attribute as well:
    
    +-----------------------+-------------------+-------------------+
    | Type attribute name   | Name              | Value             |
    +=======================+===================+===================+
    | none                  | none              |                   |
    +-----------------------+-------------------+-------------------+
    | rules                 | rules             | TERMS             |
    +-----------------------+-------------------+-------------------+
    | text_input            | text input        | TEXT_INPUT        |
    +-----------------------+-------------------+-------------------+
    | paragraph             | paragraph         | PARAGRAPH         |
    +-----------------------+-------------------+-------------------+
    | multiple_choices      | multiple choices  | MULTIPLE_CHOICES  |
    +-----------------------+-------------------+-------------------+
    | verification          | verification      | VERIFICATION      |
    +-----------------------+-------------------+-------------------+
    """
    __slots__ = ()
    
    @copy_docs(PreinstancedBase.__new__)
    def __new__(cls, value, name = None):
        if name is None:
            name = value.lower().replace('_', ' ')
        
        return PreinstancedBase.__new__(cls, value, name)
    
    
    none = P('', 'none')
    rules = P('TERMS', 'rules')
    text_input = P('TEXT_INPUT', 'text input')
    paragraph = P('PARAGRAPH', 'paragraph')
    multiple_choices = P('MULTIPLE_CHOICES', 'multiple_choices')
    verification = P('VERIFICATION', 'verification',)



__all__ = ()

from ...field_parsers import default_entity_parser_factory, nullable_string_parser_factory
from ...field_putters import default_entity_putter_factory, nullable_string_optional_putter_factory
from ...field_validators import default_entity_validator_factory, nullable_string_validator_factory
from ...user import ClientUserBase, User, ZEROUSER

from .constants import REASON_LENGTH_MAX, REASON_LENGTH_MIN

# user

parse_user = default_entity_parser_factory('user', User, default = ZEROUSER)
put_user = default_entity_putter_factory('user', ClientUserBase, ZEROUSER, force_include_internals = True)
validate_user = default_entity_validator_factory('user', ClientUserBase, default = ZEROUSER)

# reason

parse_reason = nullable_string_parser_factory('reason')
put_reason = nullable_string_optional_putter_factory('reason')
validate_reason = nullable_string_validator_factory('reason', REASON_LENGTH_MIN, REASON_LENGTH_MAX)



__all__ = ('MessageType', )

from scarletio import copy_docs, include

from ...activity import ActivityType
from ...bases import Preinstance as P, PreinstancedBase
from ...embed import EmbedType
from ...emoji.emoji.utils import _create_partial_emoji_from_fields
from ...utils import DATETIME_FORMAT_CODE, elapsed_time, sanitize_mentions, timestamp_to_datetime


Client = include('Client')


def MESSAGE_DEFAULT_CONVERTER(self):
    """
    Default message content converter.
    
    Parameters
    ----------
    self : ``Message``
        The message, what's content will be converted.
    
    Returns
    -------
    content : `None`, `str`
        The converted content if applicable. Might be empty string.
    """
    content = self.content
    if (content is not None):
        content = sanitize_mentions(content, self.guild)
    
    return content


def convert_user_add(self):
    return f'{self.author.name} added {self.mentioned_users[0].name} to the group.'

def convert_user_remove(self):
    return f'{self.author.name} removed {self.mentioned_users[0].name} from the group.'


def convert_call(self):
    call = self.call
    if call is None or not any(isinstance(user, Client) for user in call.iter_users()):
        return f'{self.author.name} started a call.'
    
    ended_at = call.ended_at
    if ended_at is None:
        # We are using a unicode as dash and not the minus sign
        return f'{self.author.name} started a call — Join the call.'
    
    # This formatting could be more accurate.
    if elapsed_time is None:
        return f'{self.author.name} started a call.'
    
    return f'{self.author.name} started a call that lasted {elapsed_time(self.created_at, ended_at)}.'


def convert_channel_name_change(self):
    return f'{self.author.name} changed the channel name: {self.content!s}'

def convert_channel_icon_change(self):
    return f'{self.author.name} changed the channel icon.'

def convert_new_pin(self):
    return f'{self.author.name_at(self.guild_id)!s} pinned a message to this channel. See all pinned messages.'


JOIN_MESSAGE_FORMATTERS = (
    (lambda name : f'{name} joined the party.'),
    (lambda name : f'{name} is here.'),
    (lambda name : f'Welcome, {name}. We hope you brought pizza.'),
    (lambda name : f'A wild {name} appeared.'),
    (lambda name : f'Swoooosh. {name} just landed.'),
    (lambda name : f'{name} just slid into the server.'),
    (lambda name : f'Big {name} showed up!'),
    (lambda name : f'Welcome {name}. Say hi!'),
    (lambda name : f'{name} hopped into the server.'),
    (lambda name : f'Everyone welcome {name}!'),
    (lambda name : f'Glad you\'re here, {name}.'),
    (lambda name : f'Good to see you, {name}.'),
    (lambda name : f'Yay you made it, {name}!'),
)

# TODO: this system changed, just pulled out the new texts from the js client source, but the calculation is bad
def convert_welcome(self):
    seed = self.id // 4194304 + 1420070400000
    formatter = JOIN_MESSAGE_FORMATTERS[seed % len(JOIN_MESSAGE_FORMATTERS)]
    user_name = self.author.name_at(self.guild_id)
    return formatter(user_name)


def convert_guild_boost(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    return f'{self.author.name} boosted {guild_name} with Nitro!'


def convert_guild_boost_tier_1(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
        
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 1!'


def convert_guild_boost_tier_2(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 2!'


def convert_guild_boost_tier_3(self):
    guild = self.channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
        
    return f'{self.author.name} boosted {guild_name} with Nitro! {guild_name} has achieved level 3!'


def convert_new_follower_channel(self):
    channel = self.channel
    guild = channel.guild
    if guild is None:
        guild_name = 'None'
    else:
        guild_name = guild.name
    
    user_name = self.author.name_at(guild)
    
    return (
        f'{user_name} has added {guild_name} #{channel.name} to this channel. '
        f'Its most important updates will show up here.'
    )


def convert_stream(self):
    user = self.author
    for activity in user.activities:
        if activity.type is ActivityType.stream:
            activity_name = activity.name
            break
    else:
        activity_name = 'Unknown'
    
    user_name = user.name_at(self.guild_id)
    
    return f'{user_name} is live! Now streaming {activity_name}'


def convert_discovery_disqualified(self):
    return (
        'This server has been removed from Server Discovery because it no longer passes all the requirements. '
        'Check `Server Settings` for more details.'
    )


def convert_discovery_requalified(self):
    return 'This server is eligible for Server Discovery again and has been automatically relisted!'


def convert_discovery_grace_period_initial_warning(self):
    return (
        'This server has failed Discovery activity requirements for 1 week. '
        'If this server fails for 4 weeks in a row, it will be automatically removed from Discovery.'
    )

def convert_discovery_grace_period_final_warning(self):
    return (
        'This server has failed Discovery activity requirements for 3 weeks in a row. '
        'If this server fails for 1 more week, it will be removed from Discovery.'
    )


def convert_thread_created(self):
    thread = self.thread
    if thread is None:
        thread_name = self.content
        if thread_name is None:
            thread_name = 'unknown'
    else:
        thread_name = thread.name
    
    return f'{self.author.name_at(self.guild_id)!s} started a thread: {thread_name!s}. See all threads.'


def convert_auto_moderation_action(self):
    embed = self.embed
    if embed is None or embed.type is not EmbedType.auto_moderation_message:
        return None
    
    content_parts = []
    should_add_next = ''
    
    description = embed.description
    if description is not None:
        content_parts.append(description)
        should_add_next = '\n'
    
    
    for field in embed.iter_fields():
        if field.name == 'keyword':
            content_parts.append(should_add_next)
            content_parts.append('Keyword: ')
            content_parts.append(field.value)
            should_add_next = ' • '
            break
    
    for field in embed.iter_fields():
        if field.name == 'rule_name':
            content_parts.append(should_add_next)
            content_parts.append('Rule: ')
            content_parts.append(field.value)
            should_add_next = ' • '
            break
    
    for field in embed.iter_fields():
        if field.name == 'timeout_duration':
            content_parts.append(should_add_next)
            content_parts.append('Time-out: ')
            content_parts.append(field.value)
            content_parts.append(' secs')
            break
    
    if content_parts:
        return ''.join(content_parts)


def convert_role_subscription_purchase(self):
    role_subscription = self.role_subscription
    if (role_subscription is None):
        return None
    
    user_name = self.author.name_at(self.guild_id)
    tier_name = role_subscription.tier_name
    total_months = role_subscription.total_months
    renewal = role_subscription.renewal
    
    guild = self.guild
    if guild is None:
        guild_name = 'unknown'
    else:
        guild_name = guild.name
    
    case = (renewal << 1) | total_months > 0
    
    content_parts = [
        user_name,
        ' ',
        ("renewed" if renewal else "joined"),
        ' **',
        tier_name,
        '** ',
        ('and has been a subscriber of', 'in their', 'and has been a subscriber of', 'as a subscriber of')[case],
        ' ',
        guild_name,
        (' for ', ' subscription', ' for ', '')[case],
    ]
    
    if total_months > 0:
        content_parts.append(str(total_months))
        content_parts.append(' month')
        if total_months != 1:
            content_parts.append('s')
    
    content_parts.append('!')
    return ''.join(content_parts)


def convert_invite_reminder(self):
    return (
        'Wondering who to invite?\n'
        'Start by inviting anyone who can help you build the server!'
    )


def convert_stage_start(self):
    return f'{self.author.name_at(self.guild_id)!s} started {self.content!s}.'


def convert_stage_end(self):
    return f'{self.author.name_at(self.guild_id)!s} ended {self.content!s}.'


def convert_stage_topic_change(self):
    return f'{self.author.name_at(self.guild_id)!s} changed the Stage topic: {self.content!s}'


def convert_stage_speaker(self):
    return f'{self.author.name_at(self.guild_id)!s} is now a speaker'


def convert_application_guild_subscription(self):
    application = self.application
    if application is None:
        application_name = 'unknown'
    else:
        application_name = application.name
    
    return f'{self.author.name_at(self.guild_id)} upgraded {application_name!s} to premium for this server! \uD83C\uDF89'


def convert_private_channel_integration_add(self):
    application = self.application
    if application is None:
        application_name = 'unknown'
    else:
        application_name = application.name
    
    return f'{self.author.name!s} added the {application_name!s} app. See our help center for more info.'


def convert_private_channel_integration_remove(self):
    application = self.application
    if application is None:
        application_name = 'unknown'
    else:
        application_name = application.name
    
    return f'{self.author.name!s} removed the {application_name} app. See our help center for more info.'


def convert_guild_incidents_enable(self):
    content_parts = [self.author.name_at(self.guild_id), ' enabled security actions for ']
    
    guild = self.guild
    if guild is None:
        guild_name = 'unknown'
    else:
        guild_name = guild.name
    content_parts.append(guild_name)
    
    content = self.content
    if content is not None:
        content_parts.append(' until ')
        content_parts.append(format(timestamp_to_datetime(content), DATETIME_FORMAT_CODE))
    
    content_parts.append('.')
    return ''.join(content_parts)


def convert_guild_incidents_disable(self):
    guild = self.guild
    if guild is None:
        guild_name = 'unknown'
    else:
        guild_name = guild.name
        
    return f'{self.author.name_at(self.guild_id)!s} disabled security actions for {guild_name!s}.'


def convert_purchase_notification(self):
    return f'Thank you,\n {self.author.name_at(self.guild_id)!s}'


def convert_poll_result(self):
    embed = self.embed
    if embed is None or embed.type is not EmbedType.poll_result:
        return None
    
    content_parts = []
    should_add_next = ''
    
    description = embed.description
    if description is not None:
        content_parts.append(description)
        should_add_next = '\n'
    
    
    for field in embed.iter_fields():
        if field.name == 'poll_question_text':
            content_parts.append(should_add_next)
            content_parts.append('The poll ')
            content_parts.append(field.value)
            content_parts.append(' has closed.')
            should_add_next = '\n'
            break
    
    for field in embed.iter_fields():
        if field.name == 'victor_answer_text':
            answer_text = field.value
            break
    else:
        answer_text = None
    
    for field in embed.iter_fields():
        if field.name == 'victor_answer_emoji_id':
            answer_emoji_id = field.value
            break
    else:
        answer_emoji_id = None
    
    for field in embed.iter_fields():
        if field.name == 'victor_answer_emoji_name':
            answer_emoji_name = field.value
            break
    else:
        answer_emoji_name = ''
    
    for field in embed.iter_fields():
        if field.name == 'victor_answer_emoji_animated':
            answer_emoji_animated = field.value == 'true'
            break
    else:
        answer_emoji_animated = False
    
    if answer_emoji_id or answer_emoji_name:
        emoji = _create_partial_emoji_from_fields(answer_emoji_name, answer_emoji_id, answer_emoji_animated)
    else:
        emoji = None
    
    text_flag = ((answer_text is not None) << 1) | (emoji is not None)
    if text_flag:
        content_parts.append(should_add_next)
        should_add_next = '\n'
        
        if text_flag & 0b01:
            content_parts.append(emoji.as_emoji)
        
        if text_flag == 0b11:
            content_parts.append(' ')
        
        if text_flag & 0b10:
            content_parts.append(answer_text)
    
    
    for field in embed.iter_fields():
        if field.name == 'victor_answer_votes':
            try:
                winning_answer_votes = int(field.value)
            except ValueError:
                winning_answer_votes = -1
            break
    else:
        winning_answer_votes = -1
    
    total_votes = -1
    
    for field in embed.iter_fields():
        if field.name == 'total_votes':
            try:
                total_votes = int(field.value)
            except ValueError:
                pass
            break
    
    if (winning_answer_votes != -1) and (total_votes != -1):
        content_parts.append(should_add_next)
        
        if (winning_answer_votes == 0) and (total_votes == 0):
            content_parts.append('There was no winner')
        
        else:
            content_parts.append('Winning answer • ')
            content_parts.append(format(winning_answer_votes * 100 / total_votes, '.0f'))
            content_parts.append('%')
    
    
    if content_parts:
        return ''.join(content_parts)


class MessageType(PreinstancedBase, value_type = int):
    """
    Represents a ``Message``'s type.
    
    Attributes
    ----------
    converter : `FunctionType`
        The converter function of the message type, what tries to convert the message's content to it's Discord side
        representation.
    
    deletable : `bool`
        Whether the message with this type can be deleted.
    
    name : `str`
        The default name of the message type.
    
    value : `int`
        The Discord side identifier value of the message type.
    
    Type Attributes
    ---------------
    Every predefined message type can be accessed as type attribute as well:
    
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | Type attribute name                       | Name                                      | Value | Converter                                         | Deletable |
    +===========================================+===========================================+=======+===================================================+===========+
    | default                                   | default                                   | 0     | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | user_add                                  | user add                                  | 1     | convert_user_add                                  | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | user_remove                               | user remove                               | 2     | convert_user_remove                               | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | call                                      | call                                      | 3     | convert_call                                      | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | channel_name_change                       | channel name change                       | 4     | convert_channel_name_change                       | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | channel_icon_change                       | channel icon change                       | 5     | convert_channel_icon_change                       | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | new_pin                                   | new pin                                   | 6     | convert_new_pin                                   | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | welcome                                   | welcome                                   | 7     | convert_welcome                                   | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost                               | guild boost                               | 8     | convert_guild_boost                               | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_1                        | guild boost tier 1                        | 9     | convert_guild_boost_tier_1                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_2                        | guild boost tier 2                        | 10    | convert_guild_boost_tier_2                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_boost_tier_3                        | guild boost tier 3                        | 11    | convert_guild_boost_tier_3                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | new_follower_channel                      | new follower channel                      | 12    | convert_new_follower_channel                      | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stream                                    | stream                                    | 13    | convert_stream                                    | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_disqualified                    | discovery disqualified                    | 14    | convert_discovery_disqualified                    | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_requalified                     | discovery requalified                     | 15    | convert_discovery_requalified                     | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_grace_period_initial_warning    | discovery grace period initial warning    | 16    | convert_discovery_grace_period_initial_warning    | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | discovery_grace_period_final_warning      | discovery grace period final warning      | 17    | convert_discovery_grace_period_final_warning      | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | thread_created                            | thread created                            | 18    | convert_thread_created                            | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | inline_reply                              | inline reply                              | 19    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | slash_command                             | slash command                             | 20    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | thread_started                            | thread started                            | 21    | MESSAGE_DEFAULT_CONVERTER                         | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | invite_reminder                           | invite reminder                           | 22    | convert_invite_reminder                           | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | context_command                           | context command                           | 23    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | auto_moderation_action                    | auto moderation action                    | 24    | convert_auto_moderation_action                    | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | role_subscription_purchase                | role subscription purchase                | 25    | convert_role_subscription_purchase                | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | interaction_premium_upsell                | interaction premium upsell                | 26    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_start                               | stage start                               | 27    | convert_stage_start                               | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_end                                 | stage end                                 | 28    | convert_stage_end                                 | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_speaker                             | stage speaker                             | 29    | convert_stage_speaker                             | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_request_to_speak                    | stage request to speak                    | 30    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | stage_topic_change                        | stage topic change                        | 31    | convert_stage_topic_change                        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | application_guild_subscription            | application guild subscription            | 32    | convert_application_guild_subscription            | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | private_channel_integration_add           | private channel integration add           | 33    | convert_private_channel_integration_add           | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | private_channel_integration_remove        | private channel integration remove        | 34    | convert_private_channel_integration_remove        | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | premium_referral                          | premium referral                          | 35    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_incidents_enable                    | guild incidents enable                    | 36    | convert_guild_incidents_enable                    | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_incidents_disable                   | guild incidents disable                   | 37    | convert_guild_incidents_disable                   | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_incidents_report_raid               | guild incidents report raid               | 38    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_incidents_report_false_alarm        | guild incidents report false alarm        | 39    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_chat_revive                         | guild chat revive                         | 40    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | custom_gift                               | custom gift                               | 41    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | guild_gaming_stats                        | guild gaming stats                        | 42    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | poll                                      | poll                                      | 43    | MESSAGE_DEFAULT_CONVERTER                         | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | purchase_notification                     | purchase notification                     | 44    | MESSAGE_DEFAULT_CONVERTER                         | false     |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | ???                                       | ???                                       | 45    | ???                                               | ???       |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    | poll_result                               | poll result                               | 46    | convert_poll_result                               | true      |
    +-------------------------------------------+-------------------------------------------+-------+---------------------------------------------------+-----------+
    """
    __slots__ = ('converter', 'deletable',)
    
    def __new__(cls, value, name = None, converter = None, deletable = True):
        """
        Creates a new message type.
        
        Parameters
        ----------
        value : `int`
            The Discord side identifier value of the message type.
        
        name : `None | str` = `None`, Optional
            The default name of the message type.
        
        converter : `None | FunctionType` = `None`, Optional
            The converter function of the message type.
        
        deletable : `bool` = `True`, Optional
            Whether the message with this type can be deleted.
        """
        if converter is None:
            converter = MESSAGE_DEFAULT_CONVERTER
        
        self = PreinstancedBase.__new__(cls, value, name)
        self.converter = converter
        self.deletable = deletable
        return self
    
    
    @copy_docs(PreinstancedBase._put_repr_parts_into)
    def _put_repr_parts_into(self, repr_parts):
        # converter
        repr_parts.append(', converter = ')
        repr_parts.append(repr(self.converter))
        
        # deletable
        repr_parts.append(', deletable = ')
        repr_parts.append(repr(self.deletable))
    
    
    # predefined
    default = P(0, 'default', MESSAGE_DEFAULT_CONVERTER, True)
    user_add = P(1, 'add user', convert_user_add, False)
    user_remove = P(2, 'remove user', convert_user_remove, False)
    call = P(3, 'call', convert_call, False)
    channel_name_change = P(4, 'channel name change', convert_channel_name_change, False)
    channel_icon_change = P(5, 'channel icon change', convert_channel_icon_change, False)
    new_pin = P(6, 'new pin', convert_new_pin, True)
    welcome = P(7, 'welcome', convert_welcome, True)
    guild_boost = P(8, 'guild boost', convert_guild_boost, True)
    guild_boost_tier_1 = P(9, 'guild boost tier 1', convert_guild_boost_tier_1, True)
    guild_boost_tier_2 = P(10, 'guild boost tier 2', convert_guild_boost_tier_2, True)
    guild_boost_tier_3 = P(11, 'guild boost tier 3', convert_guild_boost_tier_3, True)
    new_follower_channel = P(12, 'new follower channel', convert_new_follower_channel, True)
    stream = P(13, 'stream', convert_stream, False)
    discovery_disqualified = P(14, 'discovery disqualified', convert_discovery_disqualified, False)
    discovery_requalified = P(15, 'discovery requalified', convert_discovery_requalified, False)
    discovery_grace_period_initial_warning = P(
        16, 'discovery grace period initial warning', convert_discovery_grace_period_initial_warning, False
    )
    discovery_grace_period_final_warning = P(
        17, 'discovery grace period final warning', convert_discovery_grace_period_final_warning, False
    )
    thread_created = P(18, 'thread created', convert_thread_created, True)
    inline_reply = P(19, 'inline reply', MESSAGE_DEFAULT_CONVERTER, True)
    slash_command = P(20, 'slash command', MESSAGE_DEFAULT_CONVERTER, True)
    thread_started = P(21, 'thread started', MESSAGE_DEFAULT_CONVERTER, False)
    invite_reminder = P(22, 'invite reminder', convert_invite_reminder, True)
    context_command = P(23, 'context command', MESSAGE_DEFAULT_CONVERTER, True)
    auto_moderation_action = P(24, 'auto moderation action', convert_auto_moderation_action, True)
    role_subscription_purchase = P (25, 'role subscription purchase', convert_role_subscription_purchase, True)
    interaction_premium_upsell = P(26, 'interaction premium upsell', MESSAGE_DEFAULT_CONVERTER, True)
    stage_start = P(27, 'stage start', convert_stage_start, True)
    stage_end = P(28, 'stage end', convert_stage_end, True)
    stage_speaker = P(29, 'stage speaker', convert_stage_speaker, True)
    stage_request_to_speak = P(30, 'stage request to speak', MESSAGE_DEFAULT_CONVERTER, True)
    stage_topic_change = P(31, 'stage topic change', convert_stage_topic_change, True)
    application_guild_subscription = P(
        32, 'application guild subscription', convert_application_guild_subscription, True
    )
    private_channel_integration_add = P(
        33, 'private channel integration add', convert_private_channel_integration_add, True
    )
    private_channel_integration_remove = P(
        34, 'private channel integration remove', convert_private_channel_integration_remove, True
    )
    premium_referral = P(35, 'premium referral', MESSAGE_DEFAULT_CONVERTER, True)
    guild_incidents_enable = P(36, 'guild incidents enable', convert_guild_incidents_enable, False)
    guild_incidents_disable = P(37, 'guild incidents disable', convert_guild_incidents_disable, False)
    guild_incidents_report_raid = P(38, 'guild incidents report raid', MESSAGE_DEFAULT_CONVERTER, False)
    guild_incidents_report_false_alarm = P(39, 'guild incidents report false alarm', MESSAGE_DEFAULT_CONVERTER, False)
    guild_chat_revive = P(40, 'guild chat revive', MESSAGE_DEFAULT_CONVERTER, True)
    custom_gift = P(41, 'custom gif', MESSAGE_DEFAULT_CONVERTER, True)
    guild_gaming_stats = P(42, 'guild gaming stats', MESSAGE_DEFAULT_CONVERTER, True)
    poll = P(43, 'poll', MESSAGE_DEFAULT_CONVERTER, True)
    purchase_notification = P(44, 'purchase notification', convert_purchase_notification, False)
    # 45 ???
    poll_result = P(46, 'poll result', convert_poll_result, True)


GENERIC_MESSAGE_TYPES = frozenset((
    MessageType.default,
    MessageType.inline_reply,
    MessageType.slash_command,
    MessageType.thread_started,
    MessageType.context_command,
))



__all__ = ('InteractionComponent',)

from reprlib import repr as short_repr

from scarletio import RichAttributeErrorBaseType, export

from ...component import ComponentType

from .fields import (
    parse_custom_id, parse_components, parse_type, parse_value, put_custom_id,
    put_components, put_type, put_value, validate_custom_id, validate_components, validate_type,
    validate_value
)


@export
class InteractionComponent(RichAttributeErrorBaseType):
    """
    Integration component representing a parameter or a sub-command-group.
    
    Attributes
    ----------
    components : `None`, `tuple` of ``InteractionComponent``
        Nested components.
    
    custom_id : `None`, `str`
        The `.custom_id` of the represented component.
    
    type : ``ComponentType``
        The represented component's type.
    
    value : `None`, `str`
        The component's value defined by the user.
        
        > Mutually exclusive with the `components` field.
    """
    __slots__ = ('components', 'custom_id', 'type', 'value')
    
    
    def __new__(cls, *, component_type = ..., components = ..., custom_id = ..., value = ...):
        """
        Creates a new interaction component from the given keyword parameters.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Nested components.
        
        component_type : ``ComponentType``, Optional (Keyword only)
            The represented component's type.
        
        value : `None`, `str`, Optional (Keyword only)
            The component's value defined by the user.
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
        ValueError
            - If a field's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = ComponentType.none
        else:
            component_type = validate_type(component_type)
        
        # components
        if components is ...:
            components = None
        else:
            components = validate_components(components)
        
        # custom_id
        if custom_id is ...:
            custom_id = None
        else:
            custom_id = validate_custom_id(custom_id)
        
        # value
        if value is ...:
            value = None
        else:
            value = validate_value(value)
        
        # Construct
        self = object.__new__(cls)
        self.custom_id = custom_id
        self.components = components
        self.type = component_type
        self.value = value
        return self
    
    
    @classmethod
    def _create_empty(cls):
        """
        Creates an interaction component with it's default values set.
        
        Returns
        -------
        self : `instance<cls>`
        """
        self = object.__new__(cls)
        self.components = None
        self.custom_id = None
        self.type = ComponentType.none
        self.value = None
        return self
    
    
    def copy(self):
        """
        Copies the interaction component.
        
        Returns
        -------
        new : `instance<cls>`
        """
        new = object.__new__(type(self))
        components = self.components
        if (components is not None):
            components = (*(component.copy() for component in components),)
        new.components = components
        new.custom_id = self.custom_id
        new.type = self.type
        new.value = self.value
        return new
    
    
    def copy_with(self, component_type = ..., components = ..., custom_id = ..., value = ...):
        """
        Copies the interaction component with replacing the defined fields.
        
        Parameters
        ----------
        custom_id : `None`, `str`, Optional (Keyword only)
            The `.custom_id` of the represented component.
        
        components : `None`, `tuple` of ``InteractionComponent``, Optional (Keyword only)
            Nested components.
        
        component_type : ``ComponentType``, Optional (Keyword only)
            The represented component's type.
        
        value : `None`, `str`, Optional (Keyword only)
            The component's value defined by the user.
        
        Returns
        -------
        new : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a field's type is incorrect.
            - Extra or unused fields given.
        ValueError
            - If a field's value is incorrect.
        """
        # component_type
        if component_type is ...:
            component_type = self.type
        else:
            component_type = validate_type(component_type)
        
        # components
        if components is ...:
            components = self.components
            if (components is not None):
                components = (*(component.copy() for component in components),)
        else:
            components = validate_components(components)
        
        # custom_id
        if custom_id is ...:
            custom_id = self.custom_id
        else:
            custom_id = validate_custom_id(custom_id)
        
        # value
        if value is ...:
            value = self.value
        else:
            value = validate_value(value)
        
        # Construct
        self = object.__new__(type(self))
        self.custom_id = custom_id
        self.components = components
        self.type = component_type
        self.value = value
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new interaction component from the data received from Discord.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`)
            Interaction component data.
        """
        self = object.__new__(cls)
        self.components = parse_components(data)
        self.custom_id = parse_custom_id(data)
        self.type = parse_type(data)
        self.value = parse_value(data)
        return self
    
    
    def to_data(self, *, defaults = False):
        """
        Converts the interaction component into a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default field values should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        put_components(self.components, data, defaults)
        put_custom_id(self.custom_id, data, defaults)
        put_type(self.type, data, defaults)
        put_value(self.value, data, defaults)
        return data
    
    
    def __repr__(self):
        """Returns the interaction component's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        # Descriptive fields : type
        
        # type
        component_type = self.type
        if component_type is not ComponentType.none:
            repr_parts.append(' type = ')
            repr_parts.append(component_type.name)
            repr_parts.append(' ~ ')
            repr_parts.append(repr(component_type.value))
            
            field_added = True
        
        else:
            field_added = False
        
        # System fields : custom_id
        
        # custom_id
        
        custom_id = self.custom_id
        if custom_id is not None:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' custom_id = ')
            repr_parts.append(short_repr(custom_id))
        
        # Extra descriptive fields : components | value
        # components
        components = self.components
        if (components is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' components = [')
            
            index = 0
            limit = len(components)
            
            while True:
                option = components[index]
                index += 1
                repr_parts.append(repr(option))
                
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # value
        value = self.value
        if (value is not None):
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' value = ')
            repr_parts.append(repr(value))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the interaction component's hash value."""
        hash_value = 0
        
        # custom_id
        custom_id = self.custom_id
        if (custom_id is not None):
            hash_value ^= hash(custom_id)
        
        # components
        components = self.components
        if (components is not None):
            hash_value ^= (len(components) << 8)
            
            for component in components:
                hash_value ^= hash(component)
        
        # type
        hash_value ^= self.type.value
        
        # value
        value = self.value
        if (value is not None):
            hash_value ^= hash(value)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two interaction component's are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # components
        if self.components != other.components:
            return False
        
        # custom_id
        if self.custom_id != other.custom_id:
            return False
        
        # type
        if self.type is not other.type:
            return False
        
        # value
        if self.value != other.value:
            return False
        
        return True
    
    
    def iter_components(self):
        """
        Iterates over the sub-components of the interaction component.
        
        This method is an iterable generator.
        
        Yields
        ------
        component : ``InteractionComponent``
        """
        components = self.components
        if (components is not None):
            yield from components


    def iter_custom_ids_and_values(self):
        """
        Iterates over all the `custom_id`-s and values of the form submit interaction option.
        
        This method is an iterable generator.
        
        Yields
        ------
        custom_id : `str`
            The `custom_id` of a represented component.
        value : `str`
            The `value` passed by the user.
        """
        custom_id = self.custom_id
        if (custom_id is not None):
            yield custom_id, self.value
        
        for component in self.iter_components():
            yield from component.iter_custom_ids_and_values()



__all__ = ('ComponentMetadataText',)

from scarletio import copy_docs

from .base import ComponentMetadataBase
from .fields import parse_content, put_content, validate_content


class ComponentMetadataText(ComponentMetadataBase):
    """
    Represents the metadata of a text component used inside of a form.
    
    Attributes
    ----------
    content : `None`, `str`
        The content shown on the component.
    """
    __slots__ = ('content',)
    
    def __new__(cls, *, content = ...):
        """
        Creates a text component metadata.
        
        Parameters
        ----------
        content : `None`, `str`, Optional (Keyword only)
            The content shown on the component.
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # content
        if content is ...:
            content = None
        else:
            content = validate_content(content)
        
        # Construct
        self = object.__new__(cls)
        self.content = content
        return self
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_keyword_parameters)
    def from_keyword_parameters(cls, keyword_parameters):
        return cls(
            content = keyword_parameters.pop('content', ...),
        )
    
    
    @copy_docs(ComponentMetadataBase.__repr__)
    def __repr__(self):
        repr_parts = ['<', type(self).__name__]
        
        # content
        content = self.content
        if (content is not None):
            repr_parts.append(' content = ')
            repr_parts.append(repr(content))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    @copy_docs(ComponentMetadataBase.__hash__)
    def __hash__(self):
        hash_value = 0
        
        # content
        content = self.content
        if (content is not None):
            hash_value ^= hash(content)
        
        return hash_value
    
    
    @copy_docs(ComponentMetadataBase._is_equal_same_type)
    def _is_equal_same_type(self, other):
        # content
        if self.content != other.content:
            return False
        
        return True
    
    
    @classmethod
    @copy_docs(ComponentMetadataBase.from_data)
    def from_data(cls, data):
        self = object.__new__(cls)
        self.content = parse_content(data)
        return self
    
    
    @copy_docs(ComponentMetadataBase.to_data)
    def to_data(self, *, defaults = False):
        data = {}
        
        put_content(self.content, data, defaults)
        
        return data
    
    
    @copy_docs(ComponentMetadataBase.copy)
    def copy(self):
        new = object.__new__(type(self))
        
        new.content = self.content
        
        return new
    
    
    def copy_with(
        self, 
        *,
        content = ...,
    ):
        """
        Copies the text component metadata with the given fields.
        
        Parameters
        ----------
        content : `None`, `str`, Optional (Keyword only)
            The content shown on the component.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is incorrect.
        ValueError
            - If a parameter's value is incorrect.
        """
        # content
        if content is ...:
            content = self.content
        else:
            content = validate_content(content)
        
        # Construct
        new = object.__new__(type(self))
        new.content = content
        return new
    
    
    @copy_docs(ComponentMetadataBase.copy_with_keyword_parameters)
    def copy_with_keyword_parameters(self, keyword_parameters):
        return self.copy_with(
            content = keyword_parameters.pop('content', ...),
        )



__all__ = ()

from scarletio import RichAttributeErrorBaseType, export


VARIATION_SELECTOR_16_POSTFIX = '_vr16'


@export
class Unicode(RichAttributeErrorBaseType):
    """
    Represents a raw unicode emoji. These objects are further processed on the line.
    
    Attributes
    ----------
    aliases : `None`, `tuple` of `str`
        Alternative names of the emoji.
    emoticons : `None`, `tuple` of `str`
        Emoticons which the emoji represents.
    name : `str`
        The unicode's name.
    value : `str`
        Unicode value.
    variation_selector_16 : `bool`
        Whether the emoji is a variation selector 16 emoji.
    unicode_aliases : `None | tuple<str>`
        Alternative unicode strings representing the same unicode.
        These are required because some shitty systems store 32 bit characters as 2 16 bit code point parses.
    """
    __slots__ = ('aliases', 'emoticons', 'name', 'value', 'variation_selector_16', 'unicode_aliases')
    
    def __new__(cls, name, value, variation_selector_16, aliases, emoticons, unicode_aliases):
        """
        Creates a new unicode.
        
        Parameters
        ----------
        name : `str`
            The unicode's name.
        value : `str`
            Unicode value.
        variation_selector_16 : `bool`
            Whether the emoji is a variation selector 16 emoji.
        aliases : `None`, `tuple` of `str`
            Alternative names of the emoji.
        emoticons : `None`, `tuple` of `str`
            Emoticons which the emoji represents.
        unicode_aliases : `None | tuple<str>`
            Alternative unicode strings representing the same unicode.
        """
        self = object.__new__(cls)
        
        self.aliases = aliases
        self.emoticons = emoticons
        self.name = name
        self.value = value
        self.variation_selector_16 = variation_selector_16
        self.unicode_aliases = unicode_aliases
        
        return self
    
    
    def __repr__(self):
        """Returns the unicode's representation."""
        return f'<{self.__class__.__name__} name = {self.name!r}>'
    
    
    def get_system_name(self):
        """
        Returns the unicode's system name.
        
        Returns
        -------
        system_name : `str`
        """
        system_name = self.name
        
        if self.variation_selector_16:
            system_name += VARIATION_SELECTOR_16_POSTFIX
        
        return system_name
    
    
    def iter_emoticons(self):
        """
        Iterates over the unicode's emoticons.
        
        This method is an iterable generator.
        
        Yields
        ------
        emoticon : `str`
        """
        emoticons = self.emoticons
        if (emoticons is not None):
            yield from emoticons
    
    
    def iter_aliases(self):
        """
        Iterates over the unicode's aliases.
        
        This method is an iterable generator.
        
        Yields
        ------
        alias : `str`
        """
        aliases = self.aliases
        if (aliases is not None):
            yield from aliases
        
    
    def iter_alternative_names(self):
        """
        Iterates over the unicode's alternative names. Includes emoticons and aliases.
        
        This method is an iterable generator
        
        Yields
        ------
        alternative_name : `str`
        """
        yield from self.iter_emoticons()
        yield from self.iter_aliases()
    
    
    def iter_unicode_aliases(self):
        """
        Iterates over the unicode's unicode aliases.
        
        This method is an iterable generator.
        
        Yields
        ------
        unicode_alias : `str`
        """
        unicode_aliases = self.unicode_aliases
        if (unicode_aliases is not None):
            yield from unicode_aliases



__all__ = ()

from time import perf_counter

from scarletio import CancelledError, RichAttributeErrorBaseType, Task, sleep

from ..core import KOKORO

from .constants import HEARTBEAT_TIMEOUT, INTERVAL_DEFAULT, LATENCY_DEFAULT


class Kokoro(RichAttributeErrorBaseType):
    """
    A beater for Discord gateways.
    
    Attributes
    ----------
    beat_task : `None`, ``Task``
        The beat task of kokoro's gateway. Set by ``._task`` meanwhile it's gateway is connected.
    beat_waiter : `None`, ``Future``
        `.runner` task's waiter. Sleeps `.runner` between two beat.
    gateway : ``DiscordGatewayBase``
        Kokoro's owner gateway.
    interval : `float`
        The time needed to be waited between two beats.
    last_answer : `float`
        The time of the last answered heartbeat event in `perf_counter` time.
    last_send : `float`
        The last time when kokoro sent a heartbeat event in `last_send` time.
    latency : `float`
        The time interval between the last beat and the response's time.
    runner : `None | Task`
        The main keep alive task of kokoro.
    should_run : `bool`
        Whether the kokoro should run and restart itself.
    """
    __slots__ = (
        'beat_task', 'beat_waiter', 'gateway', 'interval', 'last_answer', 'last_send', 'latency', 'runner',
        'should_run'
    )
    
    def __new__(cls, gateway):
        """
        Creates a ``Kokoro`` bound to it's gateway.
        
        Parameters
        ----------
        gateway : ``DiscordGateway``, ``DiscordGatewayVoice``
            The gateway of kokoro.
        
        Returns
        -------
        kokoro : ``Kokoro``
        """
        now_ = perf_counter()
        
        self = object.__new__(cls)
        
        self.beat_task = None
        self.beat_waiter = None
        self.gateway = gateway
        self.interval = INTERVAL_DEFAULT # we set it from outside
        self.last_answer = now_
        self.last_send = now_
        self.latency = LATENCY_DEFAULT
        self.runner = None
        self.should_run = False
        
        return self
    
    
    def __repr__(self):
        """Returns kokoro's representation."""
        repr_parts = ['<', type(self).__name__,]
        
        repr_parts.append(' gateway = ')
        repr_parts.append(repr(self.gateway))
        
        if self.runner is None:
            state = 'dormant'
        else:
            state = 'beating'
        
        repr_parts.append(', state = ')
        repr_parts.append(state)
        repr_parts.append('>')
        
        return ''.join(repr_parts)


    def start(self):
        """
        Starts kokoro.
        
        Returns
        -------
        started : `bool`
            Whether the gateway was started.
        """
        # Are we running or starting?
        if (self.runner is not None):
            return False
        
        self.runner = Task(KOKORO, self._run())
        return True
    
    
    def stop(self):
        """
        Stops kokoro. Called when it's gateway is closing or is scheduled for resuming.
        """
        # Set should_run
        self.should_run = False
        
        # Stop running.
        runner = self.runner
        if (runner is not None):
            self.runner = None
            runner.cancel()
    
    
    def restart(self):
        """
        Starts kokoro's beating. Handles all the cases around the board to do so.
        """
        # First the cancel task will go through, then the start, so nothing to worry about.
        self.stop()
        self.start()
    
    
    async def _run(self):
        """
        The main keep alive task of kokoro.
        
        The control flow of the method is the following:
        
        - If `.should_run` is not `True`, break out.
        - Wait `.interval` time and set the waiting ``Future`` to `self.beat_waiter`.
            If its cancelled we already beat, continue loop.
        - If we did not get answer since last beating (what is triggered first from outside), then we stop the
            gateway. We also break out from the loop to terminate the beating state and we will wait for the
            web socket to connect again.
        - We beat one with starting `gateway.beat` as a ``Task`` and setting it to `.beat_task`. If the task is
            not completed before it's respective timeout, we stop the gateway here as well. We also break out
            from the loop to terminate the beating state and we will wait for the web socket to connect again.
            This task can also be cancelled. If cancellation occurs, we repeat the loop.
        - If the beating task is done, we update `.last_send` to the current `perf_counter` time. Repeat the loop.
        
        This method is a coroutine.
        """
        try:
            self.last_answer = perf_counter()
            gateway = self.gateway
            self.should_run = True
            
            while self.should_run:
                beat_waiter = sleep(self.interval, KOKORO)
                self.beat_waiter = beat_waiter
                try:
                    await beat_waiter.wait_for_completion()
                finally:
                    self.beat_waiter = None
                
                # Were we cancelled? If yes continue with next iteration and check whether we are should_run.
                cancellation_exception = beat_waiter.get_cancellation_exception()
                if (cancellation_exception is not None) and isinstance(cancellation_exception, CancelledError):
                    self.last_send = perf_counter()
                    continue
                
                if (self.last_answer + self.interval + HEARTBEAT_TIMEOUT) - perf_counter() <= 0.0:
                    self.should_run = False
                    Task(KOKORO, gateway.terminate())
                    break
                
                beat_task = Task(KOKORO, gateway.beat())
                beat_task.apply_timeout(HEARTBEAT_TIMEOUT)
                self.beat_task = beat_task
                try:
                    await beat_task.wait_for_completion()
                finally:
                    self.beat_task = None
                
                # Check cancellation exception. If it was timeout error we should stop beating.
                cancellation_exception = beat_task.get_cancellation_exception()
                if (cancellation_exception is None):
                    self.last_send = perf_counter()
                    continue
                
                if isinstance(cancellation_exception, TimeoutError):
                    self.should_run = False
                    Task(KOKORO, gateway.terminate())
                    break
                
                # cancelled case -> resume and check `.should_run`
                continue
        
        finally:
            # Only cancel if we have not restarted yet
            if KOKORO.current_task is self.runner:
                self.should_run = False
                self.runner = None
    
    
    def answered(self):
        """
        When gateway received answer after beating, it's gateway calls this method.
        
        This method updates `.last_answer` and `.latency` as well. Because of the updated `.last_answer`,
        ``._keep_beating`` task will know that it's gateway is still connected.
        """
        now_ = perf_counter()
        self.last_answer = now_
        self.latency = now_ - self.last_send
    
    
    async def beat_now(self):
        """
        Sends a heartbeat event now. If kokoro is not at beating state, or anywhere in between, resets itself as it just
        received a heartbeat answer right now.
        
        This method is a coroutine.
        """
        while True: # GOTO
            # case 1 : we are not running:
            if self.start():
                should_beat_now = True # True = send beat data
                break
            
            # case 2 : we wait to beat
            beat_waiter = self.beat_waiter
            if (beat_waiter is not None):
                # better skip a beat
                beat_waiter.cancel()
                should_beat_now = True # True = send beat data
                break
            
#            # case 3 : we already beat
#            beat_task = self.beat_task
#            if (beat_task is not None):
#                should_beat_now = False
#                break
            
            should_beat_now = False
            break
        
        if should_beat_now:
            await self.gateway.beat()
        else:
            self.answered()



__all__ = ('VerificationScreen', )

from scarletio import RichAttributeErrorBaseType

from ...utils import DATETIME_FORMAT_CODE

from .fields import (
    parse_description, parse_edited_at, parse_steps, put_description, put_edited_at, put_steps,
    validate_description, validate_edited_at, validate_steps
)


class VerificationScreen(RichAttributeErrorBaseType):
    """
    Represents a guild's verification screen.
    
    Attributes
    ----------
    description  : `None`, `str`
        The guild's description shown in the verification screen.
    edited_at : `None`, `datetime`
        When the last version of the screen was created.
    steps : `None`, `tuple` of ``VerificationScreenStep``
        The step in the verification screen.
    """
    __slots__ = ('edited_at', 'description', 'steps')
    
    def __new__(cls, *, edited_at = ..., description = ..., steps = ...):
        """
        Creates a new verification screen instance.
        
        Parameters
        ----------
        description  : `None`, `str`, Optional (Keyword only)
            The guild's description shown in the verification screen.
        edited_at : `None`, `datetime`, Optional (Keyword only)
            When the last version of the screen was created.
        steps : `None`, `tuple` of ``VerificationScreenStep``, Optional (Keyword only)
            The step in the verification screen.
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # edited_at
        if edited_at is ...:
            edited_at = None
        else:
            edited_at = validate_edited_at(edited_at)
        
        # description
        if description is ...:
            description = None
        else:
            description = validate_description(description)
        
        # steps
        if steps is ...:
            steps = None
        else:
            steps = validate_steps(steps)
        
        # Construct
        self = object.__new__(cls)
        self.edited_at = edited_at
        self.description = description
        self.steps = steps
        return self
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new verification screen from the given data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Verification screen data.
        """
        self = object.__new__(cls)
        self.edited_at = parse_edited_at(data)
        self.description = parse_description(data)
        self.steps = parse_steps(data)
        return self
    
    
    def to_data(self, *, defaults = False, include_internals = False):
        """
        Converts the verification screen to a json serializable object.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields should be included as well.
        
        Returns
        -------
        data : `dict` of (`str`, `object`) items
        """
        data = {}
        
        put_description(self.description, data, defaults)
        put_steps(self.steps, data, defaults)
        
        if include_internals:
            put_edited_at(self.edited_at, data, defaults)
        
        return data
    
    
    def __repr__(self):
        """Returns the verification screen's representation."""
        repr_parts = ['<', self.__class__.__name__]
        
        field_added = False
        
        # description
        description = self.description
        if (description is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' description = ')
            repr_parts.append(repr(description))
        
        # steps
        steps = self.steps
        if (steps is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' steps = [')
            
            index = 0
            limit = len(steps)
            
            while True:
                step = steps[index]
                
                repr_parts.append(repr(step))
                
                index += 1
                if index == limit:
                    break
                
                repr_parts.append(', ')
                continue
            
            repr_parts.append(']')
        
        # edited_at
        edited_at = self.edited_at
        if (edited_at is not None):
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' edited_at = ')
            repr_parts.append(format(edited_at, DATETIME_FORMAT_CODE))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    
    def __hash__(self):
        """Returns the verification screen's hash value."""
        hash_value = 0
        
        # description
        description = self.description
        if (description is not None):
            hash_value ^= hash(description)
        
        # edited_at
        # Internal field, skip
        
        # steps
        steps = self.steps
        if (steps is not None):
            hash_value ^= len(steps)
            
            for step in steps:
                hash_value ^= hash(step)
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two verification screens are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        # description
        if self.description != other.description:
            return False
        
        # steps
        if self.steps != other.steps:
            return False
        
        return True
    
    
    def copy(self):
        """
        Copies the verification screen.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        new = object.__new__(type(self))
        new.edited_at = self.edited_at
        new.description = self.description
        steps = self.steps
        if (steps is not None):
            steps = tuple(step.copy() for step in steps)
        new.steps = steps
        return new
    
    
    def copy_with(self, *, edited_at = ..., description = ..., steps = ...):
        """
        Copies the verification screen with the given fields.
        
        Parameters
        ----------
        description  : `None`, `str`, Optional (Keyword only)
            The guild's description shown in the verification screen.
        edited_at : `None`, `datetime`, Optional (Keyword only)
            When the last version of the screen was created.
        steps : `None`, `tuple` of ``VerificationScreenStep``, Optional (Keyword only)
            The step in the verification screen.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is invalid.
        ValueError
            - If a parameter's value is invalid.
        """
        # edited_at
        if edited_at is ...:
            edited_at = self.edited_at
        else:
            edited_at = validate_edited_at(edited_at)
        
        # description
        if description is ...:
            description = self.description
        else:
            description = validate_description(description)
        
        # steps
        if steps is ...:
            steps = self.steps
            if (steps is not None):
                steps = tuple(step.copy() for step in steps)
        else:
            steps = validate_steps(steps)
        
        # Construct
        new = object.__new__(type(self))
        new.edited_at = edited_at
        new.description = description
        new.steps = steps
        return new
    
    
    def iter_steps(self):
        """
        Iterates over the steps of the verification screen.
        
        This method is an iterable generator.
        
        Yields
        ------
        step : ``VerificationScreenStep``
        """
        steps = self.steps
        if (steps is not None):
            yield from steps



from .action import *
from .action_metadata import *
from .execution_event import *
from .rule import *
from .trigger_metadata import *


__all__ = (
    *action.__all__,
    *action_metadata.__all__,
    *execution_event.__all__,
    *rule.__all__,
    *trigger_metadata.__all__,
)



__all__ = (
    'Button', 'ButtonStyle', 'ChannelSelect', 'Form', 'MediaGallery', 'MentionableSelect', 'Option', 'RoleSelect',
    'Row', 'Select', 'Separator', 'SeparatorSpacingSize', 'StringSelect', 'Text', 'TextInput', 'TextInputStyle',
    'UserSelect'
)

from ...discord.component import (
    ButtonStyle, InteractionForm as Form, SeparatorSpacingSize, StringSelectOption as Option, TextInputStyle,
    create_button as Button, create_channel_select as ChannelSelect, create_media_gallery as  MediaGallery,
    create_mentionable_select as MentionableSelect, create_role_select as RoleSelect, create_row as Row,
    create_string_select as Select, create_separator as Separator, create_string_select as StringSelect,
    create_text as Text, create_text_input as TextInput, create_user_select as UserSelect
)
from ...discord.core import APPLICATION_ID_TO_CLIENT
from ...discord.exceptions import DiscordException, ERROR_CODES


async def acknowledge_component_interaction(interaction_event):
    """
    Acknowledges the given `interaction_event`. If error occurs, calls `client.events.error`.
    
    This function is a coroutine.
    
    Parameter
    ---------
    interaction_event : ``InteractionEvent``
        The component interaction event to acknowledge.
    """
    try:
        client = APPLICATION_ID_TO_CLIENT[interaction_event.application_id]
    except KeyError:
        return
        
    try:
        await client.interaction_component_acknowledge(interaction_event)
    except GeneratorExit:
        raise
    
    except BaseException as err:
        if isinstance(err, ConnectionError):
            # No Internet connection
            return
        
        if isinstance(err, DiscordException) and (err.code == ERROR_CODES.unknown_interaction):
            # We timed out, bad connection.
            return
        
        await client.events.error(client, f'acknowledge_component_interaction', err)



__all__ = ()

from .fields import (
    put_description, put_end, put_name, put_privacy_level, put_schedule, put_start,
    put_status, put_target, validate_description, validate_end, validate_name, validate_privacy_level,
    validate_schedule, validate_start, validate_status, validate_target_location, validate_target_stage,
    validate_target_voice
)


SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS = {
    'description': (validate_description, put_description),
    'end': (validate_end, put_end),
    'location': (validate_target_location, put_target),
    'name': (validate_name, put_name),
    'privacy_level': (validate_privacy_level, put_privacy_level),
    'schedule': (validate_schedule, put_schedule),
    'stage': (validate_target_stage, put_target),
    'start': (validate_start, put_start),
    'voice': (validate_target_voice, put_target),
}

SCHEDULED_EVENT_EDIT_FIELD_CONVERTERS = {
    **SCHEDULED_EVENT_CREATE_FIELD_CONVERTERS,
    'status': (validate_status, put_status),
}



__all__ = ('ActivityMetadataBase',)

from scarletio import RichAttributeErrorBaseType

from ...bases import PlaceHolder

from .flags import ActivityFlag
from .preinstanced import HangType


def _pop_empty_name(keyword_parameters):
    """
    Pops empty name from the given keyword parameters.

    Parameters
    ----------
    keyword_parameters : `dict<str, object>`
        Keyword parameters passed to ``Activity.__new__``
    """
    try:
        name = keyword_parameters['name']
    except KeyError:
        pass
    else:
        if (name is None) or (not name):
            del keyword_parameters['name']


class ActivityMetadataBase(RichAttributeErrorBaseType):
    """
    Base class for activity metadatas.
    """
    __slots__ = ()
    
    def __new__(cls):
        """
        Creates a new activity metadata.
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        return object.__new__(cls)
    
    
    @classmethod
    def from_keyword_parameters(cls, keyword_parameters):
        """
        Creates a new activity metadata from the given keyword parameters.
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters passed to ``Activity.__new__``
        
        Returns
        -------
        self : `instance<cls>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        _pop_empty_name(keyword_parameters)
        return cls()
        
    
    def __repr__(self):
        """Returns the activity metadata's representation."""
        return f'<{self.__class__.__name__}>'
    
    
    def __hash__(self):
        """Returns the activity metadata's hash value."""
        return 0
    
    
    def __eq__(self, other):
        """
        Returns whether the two activity metadatas are equal.
        """
        if type(self) is not type(other):
            return NotImplemented
        
        return self._is_equal_same_type(other)
    
    
    def _is_equal_same_type(self, other):
        """
        Returns whether the two types are equal.
        
        Helper method for ``.__eq__``
        
        Parameters
        ----------
        other : `type(self)`
            The other instance. Must be from the same type.
        
        Returns
        -------
        is_equal : `bool`
        """
        return True
    
    
    @classmethod
    def from_data(cls, data):
        """
        Creates a new activity metadata.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Activity data received from Discord.
        
        Returns
        -------
        self : `instance<cls>`
        """
        return object.__new__(cls)
    
    

    def to_data(self, *, defaults = False, include_internals = False, user = False):
        """
        Converts the activity metadata to json serializable dictionary, which can be sent with bot account to change
        activity.
        
        Parameters
        ----------
        defaults : `bool` = `False`, Optional (Keyword only)
            Whether fields with their default values should be included as well.
        include_internals : `bool` = `False`, Optional (Keyword only)
            Whether internal fields, like id-s should be present as well.
        user : `bool` = `False`, Optional (Keyword only)
            Whether not only bot compatible fields should be included.
        
        Returns
        -------
        activity_data : `dict<str, object>`
        """
        return {}
    
    
    def _update_attributes(self, data):
        """
        Updates the activity metadata by overwriting it's old attributes.
        
        Parameters
        ----------
        data : `dict` of (`str`, `object`) items
            Data received from Discord.
        """
        pass
    
    
    def _difference_update_attributes(self, data):
        """
        Updates the activity metadata and returns the changes in a `dict` of (`attribute-name`, `old-value`) items.
        
        Parameters
        ----------
        data : `dict<str, object>`
            Data received from Discord.
        
        Returns
        -------
        old_attributes : `dict<str, object>`
            All item in the returned dict is optional.
        
        Returned Data Structure
        -----------------------
        
        +-------------------+-----------------------------------+
        | Keys              | Values                            |
        +===================+===================================+
        | assets            | `None`, ``ActivityAssets``        |
        +-------------------+-----------------------------------+
        | buttons           | `None`, `tuple` of `str`          |
        +-------------------+-----------------------------------+
        | created_at        | `None`, `dateTime`                |
        +-------------------+-----------------------------------+
        | details           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | emoji             | `None`, ``Emoji``                 |
        +-------------------+-----------------------------------+
        | flags             | ``ActivityFlag``                  |
        +-------------------+-----------------------------------+
        | hang_type         | ``HangType``                      |
        +-------------------+-----------------------------------+
        | name              | `str`                             |
        +-------------------+-----------------------------------+
        | metadata          | ``ActivityMetadataBase``          |
        +-------------------+-----------------------------------+
        | party             | `None`, ``ActivityParty``         |
        +-------------------+-----------------------------------+
        | secrets           | `None`, ``ActivitySecrets``       |
        +-------------------+-----------------------------------+
        | session_id        | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | state             | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | sync_id           | `None`, `str`                     |
        +-------------------+-----------------------------------+
        | timestamps        | `None`, `ActivityTimestamps``     |
        +-------------------+-----------------------------------+
        | url               | `None`, `str`                     |
        +-------------------+-----------------------------------+
        """
        return {}
    
    
    def copy(self):
        """
        Copies the activity metadata.
        
        Returns
        -------
        new : `instance<type<self>>`
        """
        return object.__new__(type(self))
    
    
    def copy_with(self):
        """
        Copies the activity metadata with the given fields.
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        return self.copy()
    
    
    def copy_with_keyword_parameters(self, keyword_parameters):
        """
        Copies the activity metadata with the given keyword parameters
        
        Parameters
        ----------
        keyword_parameters : `dict` of (`str`, `object`) items
            Keyword parameters passed to ``Activity.__new__``
        
        Returns
        -------
        new : `instance<type<self>>`
        
        Raises
        ------
        TypeError
            - If a parameter's type is unexpected.
        ValueError
           - If an parameter's value is unexpected.
        """
        _pop_empty_name(keyword_parameters)
        return self.copy_with()
    
    
    application_id = PlaceHolder(
        0,
        """
        Returns the activity's application id.
        
        Returns
        -------
        application_id : `int`
        """
    )
    
    
    assets = PlaceHolder(
        None,
        """
        Returns the activity's assets.
        
        Returns
        -------
        assets : `None`, ``ActivityAssets``
        """
    )
    
    
    buttons = PlaceHolder(
        None,
        """
        Returns the activity's buttons.
        
        Returns
        -------
        buttons : `None`, `tuple` of `str`
        """
    )
    
    
    created_at = PlaceHolder(
        None,
        """
        Returns when the activity was created.
        
        Returns
        -------
        created_at : `None`, `Datetime`
        """
    )
    
    
    details = PlaceHolder(
        None,
        """
        Returns the activity's details.
        
        Returns
        -------
        details : `None`, `str`
        """
    )
    
    
    emoji = PlaceHolder(
        None,
        """
        Returns the emoji of the activity. If it has no emoji, then set as `None`.
        
        Returns
        -------
        emoji : `None`, ``Emoji``
        """
    )
    
    
    flags = PlaceHolder(
        ActivityFlag(),
        """
        Returns the activity's flags.
        
        Returns
        -------
        flags : ``ActivityFlag``
        """
    )
    
    
    hang_type = PlaceHolder(
        HangType.none,
        """
        Returns the activity's hang state.
        
        Returns
        -------
        hang_type : ``HangType``
        """
    )
    
    
    id = PlaceHolder(
        0,
        """
        Returns the activity's id.
        
        Returns
        -------
        id : `int`
        """
    )
    
    
    name = PlaceHolder(
        '',
        """
        Returns the activity's name.
        
        Returns
        -------
        name : `str`
        """
    )
    
    
    party = PlaceHolder(
        None,
        """
        Returns the activity's party.
        
        Returns
        -------
        party : `None`, ``ActivityParty``
        """
    )
    
    
    secrets = PlaceHolder(
        None,
        """
        Returns the activity's secrets.
        
        Returns
        -------
        secrets : `None`, ``ActivitySecrets``
        """
    )
    
    
    session_id = PlaceHolder(
        None,
        """
        Returns the activity's session identifier.
        
        Returns
        -------
        session_id : `None`, `str`
        """
    )

    
    state = PlaceHolder(
        None,
        """
        Returns the activity's state.
        
        > If the activity has ``.emoji`` it appears next to it.
        
        Returns
        -------
        state : `None`, `str`
        """
    )
    
    
    sync_id = PlaceHolder(
        None,
        """
        Returns the activity's sync identifier.
        
        Returns
        -------
        sync_id : `None`, `str`
        """
    )
    
    
    timestamps = PlaceHolder(
        None,
        """
        Returns the activity's timestamps.
        
        Returns
        -------
        timestamps : `None`, ``ActivityTimestamps``
        """
    )
    
    
    url = PlaceHolder(
        None,
        """
        Returns the activity's url.
        
        Returns
        -------
        url : `None`, `str`
        """
    )
